# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import mock
from neutron.api.v2 import attributes
from neutron.tests import fake_notifier
from neutron.tests.unit.db import test_agentschedulers_db
from neutron.tests.unit.db import test_db_base_plugin_v2
from oslo_config import cfg
import six
from webob import exc

import networking_cisco
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import (cb_constants as
    svc_constants)
from networking_cisco.plugins.cisco.common import cisco_constants as c_const
from networking_cisco.plugins.cisco.device_manager import service_vm_lib
from networking_cisco.plugins.cisco.extensions import ciscocfgagentscheduler
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support as dev_mgr_test_support)
from networking_cisco.tests.unit.cisco.device_manager import (
    test_db_device_manager)


NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION

policy_path = (os.path.abspath(networking_cisco.__path__[0]) +
               '/../etc/policy.json')

CORE_PLUGIN_KLASS = dev_mgr_test_support.CORE_PLUGIN_KLASS
HW_CATEGORY = ciscohostingdevicemanager.HARDWARE_CATEGORY
L3_CFG_HOST_A = dev_mgr_test_support.L3_CFG_HOST_A
L3_CFG_HOST_B = dev_mgr_test_support.L3_CFG_HOST_B
L3_CFG_HOST_C = dev_mgr_test_support.L3_CFG_HOST_C


class HostingDeviceCfgAgentSchedulerTestMixIn(
        test_agentschedulers_db.AgentSchedulerTestMixIn):

    def _list_hosting_devices_handled_by_cfg_agent(
            self, cfg_agent_id, expected_code=exc.HTTPOk.code,
            admin_context=True):
        path = "/agents/%s/%s.%s" % (
            cfg_agent_id, ciscocfgagentscheduler.CFG_AGENT_HOSTING_DEVICES,
            self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _list_cfg_agents_handling_hosting_device(
            self, hosting_device_id, expected_code=exc.HTTPOk.code,
            admin_context=True):
        path = "/dev_mgr/hosting_devices/%s/%s.%s" % (
            hosting_device_id,
            ciscocfgagentscheduler.HOSTING_DEVICE_CFG_AGENTS, self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _assign_hosting_device_to_cfg_agent(
            self, cfg_agent_id, hosting_device_id,
            expected_code=exc.HTTPCreated.code, admin_context=True):
        path = "/agents/%s/%s.%s" % (
            cfg_agent_id, ciscocfgagentscheduler.CFG_AGENT_HOSTING_DEVICES,
            self.fmt)
        req = self._path_create_request(
            path, {'hosting_device_id': hosting_device_id},
            admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)

    def _unassign_hosting_device_from_cfg_agent(
            self, cfg_agent_id, hosting_device_id,
            expected_code=exc.HTTPNoContent.code, admin_context=True):
        path = "/agents/%s/%s/%s.%s" % (
            cfg_agent_id, ciscocfgagentscheduler.CFG_AGENT_HOSTING_DEVICES,
            hosting_device_id, self.fmt)
        req = self._path_delete_request(path, admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)


class HostingDeviceConfigAgentSchedulerTestCaseBase(
    test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
    test_db_device_manager.DeviceManagerTestCaseMixin,
    dev_mgr_test_support.DeviceManagerTestSupportMixin,
        HostingDeviceCfgAgentSchedulerTestMixIn):

    #NOTE(bobmel): Work-around to make these unit tests to work since we
    # let the core plugin implement the device manager service.
    # The device manager service should map to hosting device extension
    svc_constants.EXT_TO_SERVICE_MAPPING[
        ciscohostingdevicemanager.HOSTING_DEVICE_MANAGER_ALIAS] = (
        c_const.DEVICE_MANAGER)
    resource_prefix_map = (test_db_device_manager.TestDeviceManagerDBPlugin
                           .resource_prefix_map)
    mock_cfg_agent_notifiers = True
    host_category = HW_CATEGORY

    def setUp(self, core_plugin=None, dm_plugin=None, ext_mgr=None):
        # Save the global RESOURCE_ATTRIBUTE_MAP
        self.saved_attr_map = {}
        for resource, attrs in six.iteritems(
                attributes.RESOURCE_ATTRIBUTE_MAP):
            self.saved_attr_map[resource] = attrs.copy()
        if not core_plugin:
            core_plugin = CORE_PLUGIN_KLASS
        service_plugins = {}
        cfg.CONF.set_override('api_extensions_path',
                              dev_mgr_test_support.extensions_path)
        if ext_mgr is None:
            ext_mgr = dev_mgr_test_support.TestDeviceManagerExtensionManager()

        super(HostingDeviceConfigAgentSchedulerTestCaseBase, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)

        # Ensure we use policy definitions from our repo
        cfg.CONF.set_override('policy_file', policy_path, 'oslo_policy')
        self.core_plugin = bc.get_plugin()
        self.plugin = self.core_plugin
        self.setup_notification_driver()

        if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
            cfg.CONF.set_override('allow_sorting', True)
        self._define_keystone_authtoken()

        self._mock_l3_admin_tenant()
        self._create_mgmt_nw_for_tests(self.fmt)
        # in unit tests we don't use keystone so we mock that session
        self.core_plugin._svc_vm_mgr_obj = service_vm_lib.ServiceVMManager(
            True, None, None, None, '', keystone_session=mock.MagicMock())
        self._mock_svc_vm_create_delete(self.core_plugin)
        if self.mock_cfg_agent_notifiers is True:
            self._mock_cfg_agent_notifier(self.plugin)
        self._other_tenant_id = dev_mgr_test_support._uuid()
        self.adminContext = bc.context.get_admin_context()

    def _setup_cfg_agents(self, host_a_active=True, host_b_active=False,
                          host_c_active=False):
        self._register_cfg_agent_states(host_a_active, host_b_active,
                                        host_c_active)
        agents = self._list(
            'agents', query_params='agent_type=%s' % c_const.AGENT_TYPE_CFG)
        self._agent_dict = {agt['host']: agt for agt in agents['agents']}


class HostingDeviceConfigAgentSchedulerTestCase(
        HostingDeviceConfigAgentSchedulerTestCaseBase):

    def _test_assign_hosting_device_to_cfg_agent(self, hosting_device,
                                                 cfg_agent_id,
                                                 id_bound_agent=None):
        hd = hosting_device['hosting_device']
        self.assertEqual(id_bound_agent, hd['cfg_agent_id'])
        self._assign_hosting_device_to_cfg_agent(cfg_agent_id, hd['id'])
        hosting_device_after = self._show('hosting_devices', hd['id'])
        hd_after = hosting_device_after['hosting_device']
        self.assertEqual(cfg_agent_id, hd_after['cfg_agent_id'])
        return hosting_device_after

    def test_hosting_device_assign_to_cfg_agent(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id)

    def test_assigned_hosting_device_assign_to_cfg_agent(self):
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                cfg_agent_id1 = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id1)
                hd = hosting_device['hosting_device']
                cfg_agent_id2 = self._agent_dict[L3_CFG_HOST_B]['id']
                self._assign_hosting_device_to_cfg_agent(
                    cfg_agent_id2, hd['id'], exc.HTTPConflict.code)
                hd_after = self._show('hosting_devices', hd['id'])[
                    'hosting_device']
                self.assertEqual(cfg_agent_id1, hd_after['cfg_agent_id'])

    def test_hosting_device_assign_to_cfg_agent_with_admin_state_down(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                self.assertIsNone(hd['cfg_agent_id'])
                id_cfg_agent_disabled = self._agent_dict[L3_CFG_HOST_A]['id']
                self._update('agents', id_cfg_agent_disabled,
                             {'agent': {'admin_state_up': False}})
                self._assign_hosting_device_to_cfg_agent(
                    id_cfg_agent_disabled, hd['id'], exc.HTTPNotFound.code)
                hd_after = self._show('hosting_devices',
                                      hd['id'])['hosting_device']
                self.assertIsNone(hd_after['cfg_agent_id'])

    def test_hosting_device_assign_to_cfg_agent_two_times(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                hd_after = self._test_assign_hosting_device_to_cfg_agent(
                    hosting_device, cfg_agent_id)
                self._test_assign_hosting_device_to_cfg_agent(hd_after,
                                                              cfg_agent_id,
                                                              cfg_agent_id)

    def test_hosting_device_assign_to_non_existing_cfg_agent(self):
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                no_cfg_agent_id = '00000000-0000-0000-0000-000000000099'
                self._assign_hosting_device_to_cfg_agent(
                    no_cfg_agent_id, hd['id'], exc.HTTPNotFound.code)
                hd_after = self._show('hosting_devices', hd['id'])[
                    'hosting_device']
                self.assertIsNone(hd_after['cfg_agent_id'])

    def test_hosting_device_assign_to_non_existing_hosting_device(self):
        self._setup_cfg_agents()
        no_hd_id = '00000000-0000-0000-0000-000000000099'
        cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
        self._assign_hosting_device_to_cfg_agent(cfg_agent_id, no_hd_id,
                                                 exc.HTTPNotFound.code)

    def test_hosting_device_unassign_from_hosting_device(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id)
                hd = hosting_device['hosting_device']
                self._unassign_hosting_device_from_cfg_agent(cfg_agent_id,
                                                             hd['id'])
                hd_final = self._show('hosting_devices',
                                      hd['id'])['hosting_device']
                self.assertIsNone(hd_final['cfg_agent_id'])

    def test_hosting_device_unassign_from_non_existing_cfg_agent(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id)
                hd = hosting_device['hosting_device']
                no_cfg_agent_id = '00000000-0000-0000-0000-000000000099'
                self._unassign_hosting_device_from_cfg_agent(
                    no_cfg_agent_id, hd['id'], exc.HTTPNotFound.code)
                hd_after = self._show('hosting_devices',
                                      hd['id'])['hosting_device']
                self.assertEqual(cfg_agent_id, hd_after['cfg_agent_id'])

    def test_hosting_device_unassign_from_non_existing_hosting_device(self):
        self._setup_cfg_agents()
        no_hd_id = '00000000-0000-0000-0000-000000000099'
        cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
        self._unassign_hosting_device_from_cfg_agent(cfg_agent_id, no_hd_id,
                                                     exc.HTTPNotFound.code)

    def test_unassigned_hosting_device_unassign_from_hosting_device(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                self.assertIsNone(hd['cfg_agent_id'])
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._unassign_hosting_device_from_cfg_agent(
                    cfg_agent_id, hd['id'], exc.HTTPNotFound.code)
                hd_after = self._show('hosting_devices',
                                      hd['id'])['hosting_device']
                self.assertIsNone(hd_after['cfg_agent_id'])

    def test_hosting_device_scheduling_policy(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as (
                    hosting_device1),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device2):
                hd1 = hosting_device1['hosting_device']
                hd2 = hosting_device2['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent_id,
                                                         hd1['id'])
                self._list_hosting_devices_handled_by_cfg_agent(cfg_agent_id)
                self._list_hosting_devices_handled_by_cfg_agent(
                    cfg_agent_id, expected_code=exc.HTTPForbidden.code,
                    admin_context=False)
                self._assign_hosting_device_to_cfg_agent(
                    cfg_agent_id, hd2['id'],
                    expected_code=exc.HTTPForbidden.code, admin_context=False)
                self._unassign_hosting_device_from_cfg_agent(
                    cfg_agent_id, hd1['id'],
                    expected_code=exc.HTTPForbidden.code, admin_context=False)
                self._unassign_hosting_device_from_cfg_agent(cfg_agent_id,
                                                             hd1['id'])
                self._list_cfg_agents_handling_hosting_device(hd1['id'])
                self._list_cfg_agents_handling_hosting_device(
                    hd1['id'], expected_code=exc.HTTPForbidden.code,
                    admin_context=False)

    def test_list_hosting_devices_by_cfg_agent(self):
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as (
                    hosting_device1),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device2),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device3):
                hd1 = hosting_device1['hosting_device']
                hd2 = hosting_device2['hosting_device']
                hd3 = hosting_device3['hosting_device']
                cfg_agent1_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent1_id,
                                                         hd1['id'])
                self._assign_hosting_device_to_cfg_agent(cfg_agent1_id,
                                                         hd2['id'])
                cfg_agent2_id = self._agent_dict[L3_CFG_HOST_B]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent2_id,
                                                         hd3['id'])
                hd_list1 = self._list_hosting_devices_handled_by_cfg_agent(
                    cfg_agent1_id)
                self.assertEqual(2, len(hd_list1['hosting_devices']))
                hd1_set = {hd1['id'], hd2['id']}
                for hd in hd_list1['hosting_devices']:
                    self.assertTrue(hd['id'] in hd1_set)
                hd_list2 = self._list_hosting_devices_handled_by_cfg_agent(
                    cfg_agent2_id)
                self.assertEqual(1, len(hd_list2['hosting_devices']))
                self.assertEqual(hd3['id'],
                                 hd_list2['hosting_devices'][0]['id'])

    def test_list_hosting_devices_by_cfg_agent_with_non_existing_cfg_agent(
            self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                missing_cfg_agent_id = '00000000-0000-0000-0000-000000000099'
                self._assign_hosting_device_to_cfg_agent(cfg_agent_id,
                                                         hd['id'])
                hd_list = self._list_hosting_devices_handled_by_cfg_agent(
                    missing_cfg_agent_id)
                self.assertEqual(0, len(hd_list['hosting_devices']))

    def test_list_cfg_agents_handling_hosting_device(self):
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent_id,
                                                         hd['id'])
                a_list = self._list_cfg_agents_handling_hosting_device(
                    hd['id'])
                self.assertEqual(1, len(a_list['agents']))
                self.assertEqual(cfg_agent_id, a_list['agents'][0]['id'])

    def test_list_cfg_agents_handling_unassigned_hosting_device(self):
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                a_list = self._list_cfg_agents_handling_hosting_device(
                    hd['id'])
                self.assertEqual(0, len(a_list['agents']))

    def test_list_cfg_agents_handling_non_existent_hosting_device(self):
        self._setup_cfg_agents()
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                hd_id_non_exist = hd['id'][:-1]
                self._list_cfg_agents_handling_hosting_device(
                    hd_id_non_exist, exc.HTTPNotFound.code)

    def test_get_cfg_agents(self):
        self._setup_cfg_agents(True, True)
        agents = self.plugin.get_cfg_agents(self.adminContext)
        self.assertEqual(len(self._agent_dict), len(agents))
        for agent in agents:
            self.assertIn(agent.host, self._agent_dict)
        id_cfg_agent_disabled = self._agent_dict[L3_CFG_HOST_A]['id']
        # admin down cfg agent should not be returned
        self._update('agents', id_cfg_agent_disabled,
                     {'agent': {'admin_state_up': False}})
        active_agents = self.plugin.get_cfg_agents(self.adminContext,
                                                   active=True)
        self.assertEqual(1, len(active_agents))
        self.assertEqual(L3_CFG_HOST_B, active_agents[0].host)
        self._update('agents', id_cfg_agent_disabled,
                     {'agent': {'admin_state_up': True}})
        # cfg agent with too old time report should not be returned
        with mock.patch(
                'networking_cisco.plugins.cisco.db.scheduler.'
                'cfg_agentschedulers_db.timeutils.'
                'is_older_than') as is_older_mock:
            is_older_mock.side_effect = [False, True]
            alive_agents = self.plugin.get_cfg_agents(self.adminContext,
                                                      active=True)
            self.assertEqual(1, len(alive_agents))
            self.assertEqual(L3_CFG_HOST_A, alive_agents[0].host)

    def test_get_cfg_agents_filtered(self):
        self._setup_cfg_agents(True, True, True)
        hosts = [L3_CFG_HOST_B, L3_CFG_HOST_C]
        agents = self.plugin.get_cfg_agents(self.adminContext,
                                            filters={'host': hosts})
        self.assertEqual(2, len(agents))
        for agent in agents:
            self.assertIn(agent.host, hosts)
        id_cfg_agent_disabled = self._agent_dict[L3_CFG_HOST_B]['id']
        # admin down agent should not be returned
        self._update('agents', id_cfg_agent_disabled,
                     {'agent': {'admin_state_up': False}})
        active_agents = self.plugin.get_cfg_agents(
            self.adminContext, active=True, filters={'host': hosts})
        self.assertEqual(1, len(active_agents))
        self.assertEqual(L3_CFG_HOST_C, active_agents[0].host)
        self._update('agents', id_cfg_agent_disabled,
                     {'agent': {'admin_state_up': True}})
        # cfg agent with too old time report should not be returned
        with mock.patch(
                'networking_cisco.plugins.cisco.db.scheduler.'
                'cfg_agentschedulers_db.timeutils.'
                'is_older_than') as is_older_mock:
            is_older_mock.side_effect = [False, True]
            alive_agents = self.plugin.get_cfg_agents(
                self.adminContext, active=True, filters={'host': hosts})
            self.assertEqual(1, len(alive_agents))
            self.assertEqual(L3_CFG_HOST_B, alive_agents[0].host)

    def _test_get_cfg_agents_for_hosting_devices(self, assignable_agents,
                                                 schedule=False,
                                                 id_disable_cfg_agent=None):
        def faked_choice(seq):
            res = seq[self._current_index]
            self._current_index += 1
            return res

        if id_disable_cfg_agent is not None:
            self._update('agents', id_disable_cfg_agent,
                         {'agent': {'admin_state_up': False}})
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as (
                    hosting_device1),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device2),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device3):
                hds = [hd['hosting_device'] for hd in (
                    hosting_device1, hosting_device2, hosting_device3)]
                hd_ids = [hd['id'] for hd in hds]
                # hosting devices initially not assigned to any cfg agent
                for hd in hds:
                    self.assertIsNone(hd['cfg_agent_id'])
                # hosting device with index 1 is to be left unassigned
                ids_hd_to_assign = [hds[0]['id'], hds[2]['id']]
                self._current_index = 0
                with mock.patch(
                    'networking_cisco.plugins.cisco.device_manager.scheduler'
                    '.hosting_device_cfg_agent_scheduler'
                        '.random.choice') as random_mock:
                    random_mock.side_effect = faked_choice
                    agents = self.plugin.get_cfg_agents_for_hosting_devices(
                        self.adminContext, ids_hd_to_assign, schedule=schedule)
                    self .assertEqual(len(assignable_agents), len(agents))
                    for agent in agents:
                        self.assertIn(agent['id'], assignable_agents)
                    hds_after_dict = {hd['id']: hd for hd in self._list(
                        'hosting_devices')['hosting_devices']}
                    hds_after = [hds_after_dict[i] for i in hd_ids]
                    if len(assignable_agents) == 0:
                        # with no assignable cfg agents none of the hosting
                        # devices should have been assigned a cfg agent
                        for hd_after in hds_after:
                            self.assertIsNone(hd_after['cfg_agent_id'])
                    else:
                        # hosting devices with index 0 and 2 should have been
                        # assigned a config agent (each one different)
                        self.assertIn(hds_after[0]['cfg_agent_id'],
                                      assignable_agents)
                        # remove agent from set to ensure the next assert
                        # can only match the other agent
                        assignable_agents.discard(hds_after[0]['cfg_agent_id'])
                        self.assertIn(hds_after[2]['cfg_agent_id'],
                                      assignable_agents)
                        # hosting device with index should still not have a
                        # cfg agent assigned
                        self.assertIsNone(hds_after[1]['cfg_agent_id'])

    def test_get_cfg_agents_for_hosting_devices(self):
        self._setup_cfg_agents(True, True, True)
        # the cfg agents that the hosting devices will be assigned to
        assignable_agents = {self._agent_dict[L3_CFG_HOST_A]['id'],
                             self._agent_dict[L3_CFG_HOST_B]['id']}
        self._test_get_cfg_agents_for_hosting_devices(assignable_agents, True)

    def test_get_cfg_agents_for_hosting_devices_no_schedule(self):
        self._setup_cfg_agents(True, True, True)
        # the hosting devices should not be assigned to any cfg agent
        assignable_agents = set()
        self._test_get_cfg_agents_for_hosting_devices(assignable_agents)

    def test_get_cfg_agents_for_hosting_devices_cfg_agent_admin_down(self):
        self._setup_cfg_agents(True, True, True)
        # the cfg agents that the hosting devices will be assigned to
        assignable_agents = {self._agent_dict[L3_CFG_HOST_A]['id'],
                             self._agent_dict[L3_CFG_HOST_C]['id']}
        # disable cfg agent on host_b
        id_disable_cfg_agent = self._agent_dict[L3_CFG_HOST_B]['id']
        self._test_get_cfg_agents_for_hosting_devices(assignable_agents, True,
                                                      id_disable_cfg_agent)

    def test_get_cfg_agents_for_hosting_devices_cfg_agent_admin_down_no_sched(
           self):
        self._setup_cfg_agents(True, True, True)
        # the hosting devices should not be assigned to any cfg agent
        assignable_agents = set()
        # disable cfg agent on host_b
        id_disable_cfg_agent = self._agent_dict[L3_CFG_HOST_B]['id']
        self._test_get_cfg_agents_for_hosting_devices(assignable_agents, False,
                                                      id_disable_cfg_agent)

    def test_get_cfg_agents_for_hosting_devices_reschedules_from_dead(self):
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id)
                # cfg agent with too old time report should not be returned
                with mock.patch(
                    'networking_cisco.plugins.cisco.db.scheduler.'
                    'cfg_agentschedulers_db.timeutils.is_older_than') as (
                        is_older_mock):
                    # the cfg agents on host_a and host_b are dead and
                    # alive, respectively
                    is_older_mock.side_effect = [True, True, False]
                    agents = self.plugin.get_cfg_agents_for_hosting_devices(
                        self.adminContext, [hd['id']], schedule=True)
                    cfg_agent_id_after = self._agent_dict[L3_CFG_HOST_B]['id']
                    # the hosting device should now be rescheduled and thereby
                    # be bound to the cfg agent on host_b
                    self.assertEqual(1, len(agents))
                    self.assertEqual(cfg_agent_id_after, agents[0]['id'])
                    hosting_device_after = self._show('hosting_devices',
                                                      hd['id'])
                    hd_after = hosting_device_after['hosting_device']
                    self.assertEqual(cfg_agent_id_after,
                                     hd_after['cfg_agent_id'])

    def test__check_config_agents_auto_adds_new_cfg_agents(self):
        self._setup_cfg_agents()
        with mock.patch(
            'networking_cisco.plugins.cisco.db.scheduler.'
            'cfg_agentschedulers_db.timeutils.'
            'is_older_than') as is_older_mock:
            # make the cfg agent appear to have recent timestamps
            is_older_mock.side_effect = [False, False]
            self.plugin._check_config_agents()
            self.assertEqual(len(self._agent_dict),
                             len(self.plugin._cfg_agent_statuses))
            agent_ids = {attrs['id'] for attrs in self._agent_dict.values()}
            for agent_id, info in six.iteritems(
                    self.plugin._cfg_agent_statuses):
                self.assertIn(agent_id, agent_ids)

    def test__check_config_agents_stops_monitoring_non_existent_cfg_agents(
            self):
        self.plugin._cfg_agent_statuses['non_existent_agent_id'] = {
            'timestamp': 'faketime'}
        with mock.patch(
            'networking_cisco.plugins.cisco.db.scheduler.'
            'cfg_agentschedulers_db.timeutils.is_older_than') as is_older_mock:
            # The non-existent cfg agent should have a too old time stamp
            is_older_mock.side_effect = [True]
            self.plugin._check_config_agents()
            # verify that the non-existent cfg agent is no longer monitored
            self.assertEqual(0, len(self.plugin._cfg_agent_statuses))

    def test__check_config_agents_dead_cfg_agent_triggers_hd_rescheduling(
            self):
        self._setup_cfg_agents()
        with mock.patch(
            'networking_cisco.plugins.cisco.db.scheduler.'
            'cfg_agentschedulers_db.timeutils.is_older_than') as is_older_mock:
            # make the cfg agent appear to have recent timestamps
            is_older_mock.side_effect = [False, True, True, True]
            with mock.patch('networking_cisco.plugins.cisco.db.scheduler.'
                            'cfg_agentschedulers_db.CfgAgentSchedulerDbMixin.'
                            '_reschedule_hosting_devices') as resched_mock:
                self.plugin._check_config_agents()
                self.assertEqual(1, len(self.plugin._cfg_agent_statuses))
                agent_id = list(self._agent_dict.values())[0]['id']
                self.assertEqual(
                    list(self.plugin._cfg_agent_statuses.keys())[0], agent_id)
                resched_mock.assert_called_once_with(mock.ANY, agent_id)

    def test__reschedule_hosting_devices_no_other_cfg_agent(self):
        self._setup_cfg_agents(True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device,
                                                              cfg_agent_id)
                hd_after = self._show('hosting_devices',
                                      hd['id'])['hosting_device']
                self.assertEqual(cfg_agent_id, hd_after['cfg_agent_id'])
                notify_mock = mock.MagicMock()
                self.plugin.agent_notifiers[c_const.AGENT_TYPE_CFG] = (
                    notify_mock)
                e_context = bc.context.get_admin_context()
                with mock.patch(
                        'networking_cisco.plugins.cisco.db.scheduler.'
                        'cfg_agentschedulers_db.timeutils.is_older_than') as (
                        is_older_mock):
                    # make the cfg agent appear to have outdated timestamps
                    is_older_mock.side_effect = [True]
                    self.plugin._reschedule_hosting_devices(e_context,
                                                            cfg_agent_id)
                    hd_final = self._show('hosting_devices',
                                          hd['id'])['hosting_device']
                    self.assertIsNone(hd_final['cfg_agent_id'])
                    self.assertEqual(0, notify_mock.call_count)

    def test__reschedule_hosting_devices_to_other_cfg_agent(self):
        random_patch = mock.patch('random.choice')
        random_mock = random_patch.start()

        def side_effect(seq):
            return seq[0]

        random_mock.side_effect = side_effect
        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(
                    template_id=hdt['id']) as hosting_device_1,\
                    self.hosting_device(
                        template_id=hdt['id']) as hosting_device_2:
                cfg_agent_id1 = self._agent_dict[L3_CFG_HOST_A]['id']
                self._test_assign_hosting_device_to_cfg_agent(hosting_device_1,
                                                              cfg_agent_id1)
                self._test_assign_hosting_device_to_cfg_agent(hosting_device_2,
                                                              cfg_agent_id1)
                hds_after = self._list('hosting_devices')['hosting_devices']
                for hd_after in hds_after:
                    self.assertEqual(cfg_agent_id1, hd_after['cfg_agent_id'])
                notify_mock = mock.MagicMock()
                self.plugin.agent_notifiers[c_const.AGENT_TYPE_CFG] = (
                    notify_mock)
                e_context = bc.context.get_admin_context()
                with mock.patch(
                        'networking_cisco.plugins.cisco.db.scheduler.'
                        'cfg_agentschedulers_db.timeutils.is_older_than') as (
                        is_older_mock):
                    # make the cfg agent appear to have outdated timestamps
                    is_older_mock.side_effect = [True, False, True, False]
                    self.plugin._reschedule_hosting_devices(e_context,
                                                            cfg_agent_id1)
                    cfg_agent_id2 = self._agent_dict[L3_CFG_HOST_B]['id']
                    hds_final = self._list(
                        'hosting_devices')['hosting_devices']
                    for hd_final in hds_final:
                        self.assertEqual(cfg_agent_id2,
                                         hd_final['cfg_agent_id'])
        assign_notify_mock = notify_mock.hosting_devices_assigned_to_cfg_agent
        hd_ids = [hd['id'] for hd in hds_final]
        assign_notify_mock.assert_called_once_with(
            mock.ANY, hd_ids, self._agent_dict[L3_CFG_HOST_B]['host'])
        random_patch.stop()


class HostingDeviceConfigAgentNotifierTestCase(
        HostingDeviceConfigAgentSchedulerTestCaseBase):

    mock_cfg_agent_notifiers = False

    def setUp(self, core_plugin=None, dm_plugin=None, ext_mgr=None):
        super(HostingDeviceConfigAgentNotifierTestCase, self).setUp(
            core_plugin, dm_plugin, ext_mgr)
        fake_notifier.reset()

    def test_hosting_device_assign_to_cfg_agent_notification(self):
        cfg_notifier = self.plugin.agent_notifiers[c_const.AGENT_TYPE_CFG]
        with mock.patch.object(cfg_notifier.client, 'prepare',
                               return_value=cfg_notifier.client) as (
            mock_prepare),\
            mock.patch.object(cfg_notifier.client, 'cast') as mock_cast,\
            self.hosting_device_template(host_category=self.host_category) as (
                hosting_device_template):
            hdt = hosting_device_template['hosting_device_template']
            self._setup_cfg_agents()
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent_id,
                                                         hd['id'])
                mock_prepare.assert_called_with(
                    server=L3_CFG_HOST_A)
                mock_cast.assert_called_with(
                    mock.ANY, 'hosting_devices_assigned_to_cfg_agent',
                    payload={'hosting_device_ids': [hd['id']]})
                notifications = fake_notifier.NOTIFICATIONS
                expected_event_type = 'agent.hosting_device.add'
                self._assert_notify(notifications, expected_event_type)

    def test_hosting_device_unassign_from_cfg_agent_notification(self):
        cfg_notifier = self.plugin.agent_notifiers[c_const.AGENT_TYPE_CFG]
        with mock.patch.object(cfg_notifier.client, 'prepare',
                               return_value=cfg_notifier.client) as (
                mock_prepare),\
                mock.patch.object(cfg_notifier.client, 'cast') as mock_cast,\
                self.hosting_device_template(
                    host_category=self.host_category) as (
                        hosting_device_template):
            hdt = hosting_device_template['hosting_device_template']
            self._setup_cfg_agents()
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                cfg_agent_id = self._agent_dict[L3_CFG_HOST_A]['id']
                self._assign_hosting_device_to_cfg_agent(cfg_agent_id,
                                                         hd['id'])
                self._unassign_hosting_device_from_cfg_agent(cfg_agent_id,
                                                             hd['id'])
                mock_prepare.assert_called_with(
                    server=L3_CFG_HOST_A)
                mock_cast.assert_called_with(
                    mock.ANY, 'hosting_devices_unassigned_from_cfg_agent',
                    payload={'hosting_device_ids': [hd['id']]})
                notifications = fake_notifier.NOTIFICATIONS
                expected_event_type = 'agent.hosting_device.remove'
                self._assert_notify(notifications, expected_event_type)

    def test_hosting_device_assign_from_cfg_agent_notification_when_schedule(
            self):
        cfg_notifier = self.plugin.agent_notifiers[c_const.AGENT_TYPE_CFG]
        with mock.patch.object(
                cfg_notifier.client, 'prepare',
                return_value=cfg_notifier.client) as mock_prepare,\
                mock.patch.object(cfg_notifier.client, 'cast') as mock_cast,\
                self.hosting_device_template(
                    host_category=self.host_category) as (
                        hosting_device_template):
            hdt = hosting_device_template['hosting_device_template']
            self._setup_cfg_agents()
            with self.hosting_device(template_id=hdt['id']) as hosting_device:
                hd = hosting_device['hosting_device']
                self.plugin.get_cfg_agents_for_hosting_devices(
                        self.adminContext, [hd['id']], schedule=True)
                mock_prepare.assert_called_with(
                    server=L3_CFG_HOST_A)
                mock_cast.assert_called_with(
                    mock.ANY, 'hosting_devices_assigned_to_cfg_agent',
                    payload={'hosting_device_ids': [hd['id']]})


class HostingDeviceToCfgAgentRandomSchedulerTestCase(
        HostingDeviceConfigAgentSchedulerTestCaseBase):

    def test_random_scheduling(self):
        random_patch = mock.patch('random.choice')
        random_mock = random_patch.start()

        def side_effect(seq):
            return seq[0]
        random_mock.side_effect = side_effect

        self._setup_cfg_agents(True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as (
                    hosting_device_1):
                hd1 = hosting_device_1['hosting_device']
                agents = self.plugin.get_cfg_agents_for_hosting_devices(
                    self.adminContext, [hd1['id']], schedule=True)
                self.assertEqual(1, len(agents))
                self.assertEqual(1, random_mock.call_count)
                with self.hosting_device(template_id=hdt['id']) as (
                        hosting_device_2):
                    hd2 = hosting_device_2['hosting_device']
                    self.plugin.get_cfg_agents_for_hosting_devices(
                        self.adminContext, [hd2['id']], schedule=True)
                    self.assertEqual(2, random_mock.call_count)


class HostingDeviceToCfgAgentStingySchedulerTestCase(
        HostingDeviceConfigAgentSchedulerTestCaseBase):

    def setUp(self, core_plugin=None, dm_plugin=None, ext_mgr=None):
        cfg.CONF.set_override('configuration_agent_scheduler_driver',
                              'networking_cisco.plugins.cisco.device_manager.'
                              'scheduler.hosting_device_cfg_agent_scheduler.'
                              'StingyHostingDeviceCfgAgentScheduler',
                              'general')
        super(HostingDeviceToCfgAgentStingySchedulerTestCase, self).setUp(
            core_plugin=core_plugin, dm_plugin=dm_plugin, ext_mgr=ext_mgr)

    def test_stingy_scheduling(self):
        self._setup_cfg_agents(True, True, True)
        with self.hosting_device_template(
                host_category=self.host_category) as hosting_device_template:
            hdt = hosting_device_template['hosting_device_template']
            with self.hosting_device(template_id=hdt['id']) as (
                    hosting_device_1),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device_2),\
                    self.hosting_device(template_id=hdt['id']) as (
                        hosting_device_3):
                hd1 = hosting_device_1['hosting_device']
                hd2 = hosting_device_2['hosting_device']
                hd3 = hosting_device_3['hosting_device']
                hd_ids = [hd1['id'], hd2['id'], hd3['id']]
                agents = self.plugin.get_cfg_agents_for_hosting_devices(
                    self.adminContext, hd_ids, schedule=True)
                self.assertEqual(3, len(agents))
                self.assertNotEqual(agents[0]['id'], agents[1]['id'])
                self.assertNotEqual(agents[0]['id'], agents[2]['id'])
                self.assertNotEqual(agents[1]['id'], agents[2]['id'])



# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import socket
import ssl

from oslo_config import cfg

from neutron.tests import base

from networking_cisco.ml2_drivers.ucsm import ucs_ssl
from networking_cisco.ml2_drivers.ucsm import ucsm_network_driver

from ucsmsdk import ucsdriver


class TestCiscoUcsmSSL(base.BaseTestCase):

    """Unit tests for SSL overrides."""

    def test_SSLContext_verify_true(self):
        cfg.CONF.set_override("ucsm_https_verify", True,
                              group="ml2_cisco_ucsm")
        context = ucs_ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        self.assertEqual(context.verify_mode, ssl.CERT_REQUIRED)

    def test_SSLContext_verify_false(self):
        cfg.CONF.set_override("ucsm_https_verify", False,
                              group="ml2_cisco_ucsm")
        context = ucs_ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        self.assertEqual(context.verify_mode, ssl.CERT_NONE)

    def test_wrap_socket_verify_true(self):
        cfg.CONF.set_override("ucsm_https_verify", True,
                              group="ml2_cisco_ucsm")
        sock = socket.socket()
        context = ucs_ssl.wrap_socket(sock).context
        self.assertEqual(context.verify_mode, ssl.CERT_REQUIRED)

    def test_wrap_socket_verify_false(self):
        cfg.CONF.set_override("ucsm_https_verify", False,
                              group="ml2_cisco_ucsm")
        sock = socket.socket()
        context = ucs_ssl.wrap_socket(sock).context
        self.assertEqual(context.verify_mode, ssl.CERT_NONE)

    def test_wrap_socket_verify_false_cert_reqs_true(self):
        cfg.CONF.set_override("ucsm_https_verify", False,
                              group="ml2_cisco_ucsm")
        sock = socket.socket()
        context = ucs_ssl.wrap_socket(sock,
                                      cert_reqs=ssl.CERT_REQUIRED).context
        self.assertNotEqual(context.verify_mode, ssl.CERT_REQUIRED)

    def test_wrap_socket_verify_true_cert_reqs_false(self):
        cfg.CONF.set_override("ucsm_https_verify", True,
                              group="ml2_cisco_ucsm")
        sock = socket.socket()
        context = ucs_ssl.wrap_socket(sock,
                                      cert_reqs=ssl.CERT_NONE).context
        self.assertNotEqual(context.verify_mode, ssl.CERT_NONE)


class TestUcsmsdkPatch(base.BaseTestCase):

    """Unit tests for Cisco ML2 UCS Manager SSL override for ucsmsdk."""

    # Test monkey patched ssl lib gets loaded
    @mock.patch.object(ucsm_network_driver.CiscoUcsmDriver,
                       "_create_host_and_sp_dicts_from_config")
    def test_ucsmsdk_ssl_monkey_patch(self, mock_create_host):
        network_driver = ucsm_network_driver.CiscoUcsmDriver()
        self.assertNotEqual(ucsdriver.ssl, ucs_ssl)

        network_driver._import_ucsmsdk()
        self.assertEqual(ucsdriver.ssl, ucs_ssl)

    @mock.patch.object(ucsdriver.socket, "create_connection")
    def test_ucsmsdk_default_behaviour_of_ssl_cert_checking(self,
                                                            mocked_socket):
        mocked_socket.side_effect = lambda *_: socket.socket()

        # First connection method
        tls_context = ucsdriver.TLSConnection('127.0.0.1', port=7777)
        tls_context.connect()

        self.assertEqual(tls_context.sock.context.verify_mode, ssl.CERT_NONE)

        # Second connection method
        tls1_context = ucsdriver.TLS1Connection('127.0.0.1', port=7777)
        tls1_context.connect()

        self.assertEqual(tls1_context.sock.context.verify_mode, ssl.CERT_NONE)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import socket

from keystoneauth1.identity import generic
from keystoneauth1 import session
from keystoneclient import client as k_client
from neutronclient.v2_0 import client as n_client

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import rpc


LOG = logging.getLogger(__name__)


class EventsHandler(object):

    """This class defines methods to listen and process events."""

    def __init__(self, ser_name, pqueue, c_pri, d_pri):
        self._service = None
        self._service_name = ser_name
        self._clients = {}
        self._nclient = None
        self._pq = pqueue
        self._create_pri = c_pri
        self._update_pri = c_pri
        self._delete_pri = d_pri
        self._cfg = config.CiscoDFAConfig().cfg
        self._q_agent = constants.DFA_AGENT_QUEUE
        self._url = self._cfg.dfa_rpc.transport_url
        dfaq = self._cfg.dfa_notify.cisco_dfa_notify_queue % (
            {'service_name': ser_name})
        notify_queue = dfaq
        self._notify_queue = dfaq if dfaq in notify_queue else None
        # Setup notification listener for the events.
        self._setup_notification_listener(self._notify_queue, self._url)

        if ser_name != 'keystone':
            return
        user = self._cfg.keystone_authtoken.username
        project = self._cfg.keystone_authtoken.project_name
        passwd = self._cfg.keystone_authtoken.password
        url = self._cfg.keystone_authtoken.auth_url
        u_domain = self._cfg.keystone_authtoken.user_domain_name
        p_domain = self._cfg.keystone_authtoken.project_domain_name

        auth = generic.Password(auth_url=url,
                                username=user,
                                password=passwd,
                                project_name=project,
                                project_domain_name=p_domain,
                                user_domain_name=u_domain)
        sess = session.Session(auth=auth)

        self._service = k_client.Client(session=sess)

    @property
    def nclient(self):
        if self._nclient:
            return self._nclient
        user = self._cfg.neutron.username
        project = self._cfg.neutron.project_name
        passwd = self._cfg.neutron.password
        url = self._cfg.neutron.auth_url
        u_domain = self._cfg.keystone_authtoken.user_domain_name
        p_domain = self._cfg.keystone_authtoken.project_domain_name
        auth = generic.Password(auth_url=url,
                                username=user,
                                password=passwd,
                                project_name=project,
                                project_domain_name=p_domain,
                                user_domain_name=u_domain)
        sess = session.Session(auth=auth)
        self._nclient = n_client.Client(session=sess)
        return self._nclient

    def _setup_notification_listener(self, topic_name, url):
        """Setup notification listener for a service."""

        self.notify_listener = rpc.DfaNotifcationListener(
            topic_name, url, rpc.DfaNotificationEndpoints(self))

    def start(self):
        if self.notify_listener:
            self.notify_listener.start()

    def wait(self):
        if self.notify_listener:
            self.notify_listener.wait()

    def create_rpc_client(self, thishost):
        clnt = self._clients.get(thishost)
        if clnt is None:
            try:
                host_ip = socket.gethostbyname(thishost)
            except socket.gaierror:
                LOG.error('Invalid host name for agent: %s', thishost)
            else:
                clnt = rpc.DfaRpcClient(self._url,
                                        '_'.join((self._q_agent, thishost)),
                                        exchange=constants.DFA_EXCHANGE)
                self._clients[thishost] = clnt
                LOG.debug('Created client for agent: %(host)s:%(ip)s',
                          {'host': thishost, 'ip': host_ip})

    def callback(self, timestamp, event_type, payload):
        """Callback method for processing events in notification queue.

        :param timestamp: time the message is received.
        :param event_type: event type in the notification queue such as
                           identity.project.created, identity.project.deleted.
        :param payload: Contains information of an event
        """
        try:
            data = (event_type, payload)
            LOG.debug('RX NOTIFICATION ==>\nevent_type: %(event)s, '
                      'payload: %(payload)s\n', (
                          {'event': event_type, 'payload': payload}))
            if 'create' in event_type:
                pri = self._create_pri
            elif 'delete' in event_type:
                pri = self._delete_pri
            elif 'update' in event_type:
                pri = self._update_pri
            else:
                pri = self._delete_pri
            self._pq.put((pri, timestamp, data))

        except Exception as exc:
            LOG.exception('Error: %(err)s for event %(event)s',
                          {'err': str(exc), 'event': event_type})

    def event_handler(self):
        """Wait on queue for listening to the events."""

        if not self._notify_queue:
            LOG.error('event_handler: no notification queue for %s',
                      self._service_name)
            return

        LOG.debug('calling event handler for %s', self)
        self.start()
        self.wait()

    def send_vm_info(self, thishost, msg):
        clnt = self._clients.get(thishost)
        if clnt is None:
            LOG.debug("send_vm_info: Agent on %s is not active.", thishost)
            return

        context = {}
        thismsg = clnt.make_msg('send_vm_info', context, msg=msg)
        resp = clnt.call(thismsg)
        LOG.debug("send_vm_info: resp = %s", resp)

    def update_ip_rule(self, thishost, msg):
        clnt = self._clients.get(thishost)
        if clnt is None:
            LOG.debug("update_ip_rule: Agent on %s is not active.", thishost)
            return

        context = {}
        thismsg = clnt.make_msg('update_ip_rule', context, msg=msg)
        resp = clnt.call(thismsg)
        LOG.debug("update_ip_rule: resp = %s", resp)

    def send_msg_to_agent(self, thishost, msg_type, msg):
        clnt = self._clients.get(thishost)
        if clnt is None:
            LOG.debug("send_msg_to_agent: Agent on %s is not active.",
                      thishost)
            return

        context = {'type': msg_type}
        thismsg = clnt.make_msg('send_msg_to_agent', context, msg=msg)
        resp = clnt.call(thismsg)
        LOG.debug("send_msg_to_agent: resp = %s", resp)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_log import log as logging

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent.service_helpers import (
    routing_svc_helper as helper)
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.extensions import routerrole


ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR
LOG = logging.getLogger(__name__)


class RoutingServiceHelperAci(helper.RoutingServiceHelper):

    def __init__(self, host, conf, cfg_agent):
        super(RoutingServiceHelperAci, self).__init__(
            host, conf, cfg_agent)
        self._router_ids_by_vrf = {}
        self._router_ids_by_vrf_and_ext_net = {}

    def _is_global_router(self, ri):
        return (ri.router.get(ROUTER_ROLE_ATTR) ==
            c_constants.ROUTER_ROLE_GLOBAL)

    def _process_new_ports(self, ri, new_ports, ex_gw_port, list_port_ids_up,
                           change_details):
        # Only add internal networks if we have an
        # external gateway -- otherwise we have no parameters
        # to use to configure the interface (e.g. VRF, IP, etc.)
        if self._is_global_router(ri):
            super(RoutingServiceHelperAci, self)._process_new_ports_global(
                ri, new_ports, ex_gw_port, list_port_ids_up)
        elif ex_gw_port:
            super(RoutingServiceHelperAci, self)._process_new_ports(
                ri, new_ports, ex_gw_port, list_port_ids_up, change_details)

    def _process_old_ports(self, ri, old_ports, ex_gw_port, change_details):
        gw_port = ex_gw_port or ri.ex_gw_port
        for p in old_ports:
            LOG.debug("++ removing port id = %s (gw = %s)" %
                      (p['id'], gw_port))
            p['change_details'] = change_details[p['network_id']]
            former_ports_on_network = p['change_details']['former_ports']
            if len(former_ports_on_network) > 1:
                # port p is one of at least two ports on the network
                # so we must deconfigure it accordingly
                is_primary = former_ports_on_network[0]['id'] == p['id']
            else:
                # only one port on network so is must be the primary
                is_primary = True
            self._set_subnet_info(p, p['subnets'][0]['id'], is_primary)
            # We can only clear the port if we stil have all
            # the relevant information (VRF and external network
            # parameters), which come from the GW port. Go ahead
            # and remove the interface from our internal state
            if gw_port or self._is_global_router(ri):
                self._internal_network_removed(ri, p, gw_port)
            ri.internal_ports.remove(p)

    def _enable_disable_ports(self, ri, ex_gw_port, internal_ports):
        pass

    def _process_gateway_set(self, ri, ex_gw_port, list_port_ids_up):
        super(RoutingServiceHelperAci,
              self)._process_gateway_set(ri, ex_gw_port, list_port_ids_up)
        # transitioned -- go enable any interfaces
        interfaces = ri.router.get(bc.constants.INTERFACE_KEY, [])
        new_ports = [p for p in interfaces
                     if (p['admin_state_up'] and
                         p not in ri.internal_ports)]
        router_role = ri.router[routerrole.ROUTER_ROLE_ATTR]
        new_ports_on_networks = self._get_ports_on_networks(router_role,
                                                            new_ports)
        current_ports_on_networks = self._get_ports_on_networks(router_role,
                                                                interfaces)
        former_ports_on_networks = (
            self._get_ports_on_networks(router_role, ri.internal_ports))
        change_details = {
            nw_id: {'new_ports': new_ports_on_networks.get(nw_id, []),
                    'current_ports': current_ports_on_networks.get(nw_id, []),
                    'old_ports': [],
                    'former_ports': former_ports_on_networks.get(nw_id, [])}
            for nw_id in set(current_ports_on_networks.keys()) |
            set(former_ports_on_networks.keys())}
        self._process_new_ports(ri, new_ports, ex_gw_port, list_port_ids_up,
                                change_details)

    def _process_gateway_cleared(self, ri, ex_gw_port):
        super(RoutingServiceHelperAci,
              self)._process_gateway_cleared(ri, ex_gw_port)

        # remove the internal networks at this time,
        # while the gateway information is still available
        # (has VRF network parameters)
        del_ports = copy.copy(ri.internal_ports)
        router_role = ri.router[routerrole.ROUTER_ROLE_ATTR]
        former_ports_on_networks = self._get_ports_on_networks(router_role,
                                                               del_ports)
        old_ports_on_networks = self._get_ports_on_networks(router_role,
                                                            del_ports)
        change_details = {
            nw_id: {'new_ports': [],
                    'current_ports': [],
                    'old_ports': old_ports_on_networks.get(nw_id, []),
                    'former_ports': former_ports_on_networks.get(nw_id, [])}
            for nw_id in former_ports_on_networks.keys()}
        self._process_old_ports(ri, del_ports, ex_gw_port, change_details)

    def _add_rid_to_vrf_list(self, ri):
        """Add router ID to a VRF list.

        In order to properly manage VRFs in the ASR, their
        usage has to be tracked. VRFs are provided with neutron
        router objects in their hosting_info fields of the gateway ports.
        This means that the VRF is only available when the gateway port
        of the router is set. VRFs can span routers, and even OpenStack
        tenants, so lists of routers that belong to the same VRF are
        kept in a dictionary, with the VRF name as the key.
        """
        if ri.ex_gw_port or ri.router.get('gw_port'):
            driver = self.driver_manager.get_driver(ri.id)
            vrf_name = driver._get_vrf_name(ri)
            if not vrf_name:
                return
            if not self._router_ids_by_vrf.get(vrf_name):
                LOG.debug("++ CREATING VRF %s" % vrf_name)
                driver._do_create_vrf(vrf_name)
            self._router_ids_by_vrf.setdefault(vrf_name, set()).add(
                ri.router['id'])

    def _remove_rid_from_vrf_list(self, ri):
        """Remove router ID from a VRF list.

        This removes a router from the list of routers that's kept
        in a map, using a VRF ID as the key. If the VRF exists, the
        router is removed from the list if it's present. If the last
        router in the list is removed, then the driver's method to
        remove the VRF is called and the map entry for that
        VRF is deleted.
        """
        if ri.ex_gw_port or ri.router.get('gw_port'):
            driver = self.driver_manager.get_driver(ri.id)
            vrf_name = driver._get_vrf_name(ri)
            if self._router_ids_by_vrf.get(vrf_name) and (
                    ri.router['id'] in self._router_ids_by_vrf[vrf_name]):
                self._router_ids_by_vrf[vrf_name].remove(ri.router['id'])
                # If this is the last router in a VRF, then we can safely
                # delete the VRF from the router config (handled by the driver)
                if not self._router_ids_by_vrf.get(vrf_name):
                    LOG.debug("++ REMOVING VRF %s" % vrf_name)
                    driver._remove_vrf(ri)
                    del self._router_ids_by_vrf[vrf_name]

    def _internal_network_added(self, ri, port, ex_gw_port):
        super(RoutingServiceHelperAci, self)._internal_network_added(
            ri, port, ex_gw_port)
        driver = self.driver_manager.get_driver(ri.id)
        vrf_name = driver._get_vrf_name(ri)
        net_name = ex_gw_port['hosting_info'].get('network_name')
        self._router_ids_by_vrf_and_ext_net.setdefault(
            vrf_name, {}).setdefault(net_name, set()).add(ri.router['id'])

    def _internal_network_removed(self, ri, port, ex_gw_port):
        """Remove an internal router port

        Check to see if this is the last port to be removed for
        a given network scoped by a VRF (note: there can be
        different mappings between VRFs and networks -- 1-to-1,
        1-to-n, n-to-1, n-to-n -- depending on the configuration
        and workflow used). If it is the last port, set the flag
        indicating that the internal sub-interface for that netowrk
        on the ASR should be deleted
        """
        itfc_deleted = False
        driver = self.driver_manager.get_driver(ri.id)
        vrf_name = driver._get_vrf_name(ri)
        network_name = ex_gw_port['hosting_info'].get('network_name')
        if self._router_ids_by_vrf_and_ext_net.get(
            vrf_name, {}).get(network_name) and (
                ri.router['id'] in
                self._router_ids_by_vrf_and_ext_net[vrf_name][network_name]):
            # If this is the last port for this neutron router,
            # then remove this router from the list
            if len(ri.internal_ports) == 1 and port in ri.internal_ports:
                self._router_ids_by_vrf_and_ext_net[
                    vrf_name][network_name].remove(ri.router['id'])

                # Check if any other routers in this VRF have this network,
                # and if not, set the flag to remove the interface
                if not self._router_ids_by_vrf_and_ext_net[vrf_name].get(
                        network_name):
                    LOG.debug("++ REMOVING NETWORK %s" % network_name)
                    itfc_deleted = True
                    del self._router_ids_by_vrf_and_ext_net[
                        vrf_name][network_name]
                    if not self._router_ids_by_vrf_and_ext_net.get(vrf_name):
                        del self._router_ids_by_vrf_and_ext_net[vrf_name]

        driver.internal_network_removed(ri, port,
                                        itfc_deleted=itfc_deleted)
        if ri.snat_enabled and ex_gw_port:
            driver.disable_internal_network_NAT(ri, port, ex_gw_port,
                                                itfc_deleted=itfc_deleted)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import logging
import netaddr
import re
import time
import xml.etree.ElementTree as ET

from oslo_config import cfg
from oslo_utils import importutils

from networking_cisco.plugins.cisco.cfg_agent import cfg_exceptions as cfg_exc
from networking_cisco.plugins.cisco.cfg_agent.device_drivers import (
    devicedriver_api)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.iosxe import (
    cisco_iosxe_snippets as snippets)
from networking_cisco.plugins.cisco.common.htparser import HTParser
from networking_cisco.plugins.cisco.extensions import ha

ncclient = importutils.try_import('ncclient')
manager = importutils.try_import('ncclient.manager')

LOG = logging.getLogger(__name__)
logging.getLogger('ncclient.transport.session').setLevel(logging.INFO)


class IosXeRoutingDriver(devicedriver_api.RoutingDriverBase):
    """Generic IOS XE Routing Driver.

    This driver encapsulates the configuration logic via NETCONF protocol to
    configure a generic (IOS-XE based) device for implementing
    Neutron L3 services. These services include routing, NAT and floating
    IPs (as per Neutron terminology).
    """

    DEV_NAME_LEN = 14

    def __init__(self, **device_params):
        try:
            self._host_ip = device_params['management_ip_address']
            self._host_ssh_port = device_params['protocol_port']
            credentials = device_params.get('credentials', {})
            self._username = credentials.get('user_name')
            self._password = credentials.get('password')
            self._timeout = (device_params.get('timeout') or
                             cfg.CONF.cfg_agent.device_connection_timeout)
            self._ncc_connection = None
            self._itfcs_enabled = False
        except KeyError as e:
            LOG.error("Missing device parameter:%s. Aborting "
                      "IosXeRoutingDriver initialization", e)
            raise cfg_exc.InitializationException()

    ###### Public Functions ########
    def router_added(self, ri):
        self._create_vrf(ri)

    def router_removed(self, ri):
        self._remove_vrf(ri)

    def internal_network_added(self, ri, port):
        self._create_sub_interface(ri, port)
        if port.get(ha.HA_INFO) is not None and ri.get(ha.ENABLED, False):
            self._add_ha(ri, port)

    def internal_network_removed(self, ri, port):
        self._remove_sub_interface(port)

    def external_gateway_added(self, ri, ext_gw_port):
        self._create_sub_interface(ri, ext_gw_port)
        # set default route via this network's gateway ip
        self._add_default_route(ri, ext_gw_port)

    def external_gateway_removed(self, ri, ext_gw_port):
        # remove default route via this network's gateway ip
        self._remove_default_route(ri, ext_gw_port)
        # finally, remove external network sub-interface
        self._remove_sub_interface(ext_gw_port)

    def enable_internal_network_NAT(self, ri, port, ext_gw_port):
        self._add_internal_nw_nat_rules(ri, port, ext_gw_port)

    def disable_internal_network_NAT(self, ri, port, ext_gw_port):
        self._remove_internal_nw_nat_rules(ri, [port], ext_gw_port)

    def floating_ip_added(self, ri, ext_gw_port, floating_ip, fixed_ip):
        self._add_floating_ip(ri, floating_ip, fixed_ip)

    def floating_ip_removed(self, ri, ext_gw_port, floating_ip, fixed_ip):
        self._remove_floating_ip(ri, ext_gw_port, floating_ip, fixed_ip)

    def routes_updated(self, ri, action, route):
        self._update_routing_table(ri, action, route)

    def clear_connection(self):
        self._ncc_connection = None

    def cleanup_invalid_cfg(self, hd, routers):
        # at this point nothing to be done
        return

    def get_configuration(self):
        return self._get_running_config(split=False)

    ##### Internal Functions  ####

    def _create_sub_interface(self, ri, port):
        vrf_name = self._get_vrf_name(ri)
        ip_cidr = port['ip_cidr']
        net_mask = netaddr.IPNetwork(ip_cidr).netmask
        gateway_ip = ip_cidr.split('/')[0]
        sub_interface = self._get_interface_name_from_hosting_port(port)
        vlan = self._get_interface_vlan_from_hosting_port(port)
        self._do_create_sub_interface(sub_interface, vlan, vrf_name,
                                      gateway_ip, net_mask)

    def _remove_sub_interface(self, port):
        sub_interface = self._get_interface_name_from_hosting_port(port)
        self._do_remove_sub_interface(sub_interface)

    def _add_ha(self, ri, port):
        func_dict = {
            ha.HA_HSRP: self._add_ha_hsrp,
            ha.HA_VRRP: self._add_ha_vrrp,
            ha.HA_GLBP: self._add_ha_gblp
        }
        # invoke the right function for the ha type
        func_dict[ri[ha.DETAILS][ha.TYPE]](self, ri, port)

    def _add_ha_hsrp(self, ri, port):
        priority = ri[ha.DETAILS][ha.PRIORITY]
        port_ha_info = port[ha.HA_INFO]
        group = port_ha_info['group']
        ip = port_ha_info['ha_port']['fixed_ips'][0]['ip_address']
        if ip and group and priority:
            vrf_name = self._get_vrf_name(ri)
            sub_interface = self._get_interface_name_from_hosting_port(port)
            self._do_add_ha_hsrp(sub_interface, vrf_name, priority, group, ip)

    def _add_ha_vrrp(self, ri, port):
        raise NotImplementedError()

    def _add_ha_gblp(self, ri, port):
        raise NotImplementedError()

    def _remove_ha(self, ri, port):
        pass

    def _get_acl_name_from_vlan(self, vlan):
        return "acl_%s" % vlan

    def _add_internal_nw_nat_rules(self, ri, port, ext_port):
        vrf_name = self._get_vrf_name(ri)
        in_vlan = self._get_interface_vlan_from_hosting_port(port)
        acl_no = self._get_acl_name_from_vlan(in_vlan)
        internal_cidr = port['ip_cidr']
        internal_net = netaddr.IPNetwork(internal_cidr).network
        net_mask = netaddr.IPNetwork(internal_cidr).hostmask
        inner_itfc = self._get_interface_name_from_hosting_port(port)
        outer_itfc = self._get_interface_name_from_hosting_port(ext_port)
        self._nat_rules_for_internet_access(acl_no, internal_net,
                                            net_mask, inner_itfc,
                                            outer_itfc, vrf_name)

    def _remove_internal_nw_nat_rules(self, ri, ports, ext_port):
        acls = []
        # first disable nat in all inner ports
        for port in ports:
            in_itfc_name = self._get_interface_name_from_hosting_port(port)
            inner_vlan = self._get_interface_vlan_from_hosting_port(port)
            acls.append(self._get_acl_name_from_vlan(inner_vlan))
            self._remove_interface_nat(in_itfc_name, 'inside')
        # wait for two seconds
        LOG.debug("Sleep for 2 seconds before clearing NAT rules")
        time.sleep(2)
        # clear the NAT translation table
        self._remove_dyn_nat_translations()
        # remove dynamic nat rules and acls
        vrf_name = self._get_vrf_name(ri)
        ext_itfc_name = self._get_interface_name_from_hosting_port(ext_port)
        for acl in acls:
            self._remove_dyn_nat_rule(acl, ext_itfc_name, vrf_name)

    def _add_default_route(self, ri, ext_gw_port):
        ext_gw_ip = ext_gw_port['subnet']['gateway_ip']
        if ext_gw_ip:
            vrf_name = self._get_vrf_name(ri)
            conf_str = snippets.DEFAULT_ROUTE_CFG % (vrf_name, ext_gw_ip)
            if not self._cfg_exists(conf_str):
                conf_str = snippets.SET_DEFAULT_ROUTE % (vrf_name, ext_gw_ip)
                self._edit_running_config(conf_str, 'SET_DEFAULT_ROUTE')

    def _remove_default_route(self, ri, ext_gw_port):
        ext_gw_ip = ext_gw_port['subnet']['gateway_ip']
        if ext_gw_ip:
            vrf_name = self._get_vrf_name(ri)
            conf_str = snippets.DEFAULT_ROUTE_CFG % (vrf_name, ext_gw_ip)
            if self._cfg_exists(conf_str):
                conf_str = snippets.REMOVE_DEFAULT_ROUTE % (vrf_name,
                                                            ext_gw_ip)
                self._edit_running_config(conf_str, 'REMOVE_DEFAULT_ROUTE')

    def _add_floating_ip(self, ri, floating_ip, fixed_ip):
        vrf_name = self._get_vrf_name(ri)
        self._do_add_floating_ip(floating_ip, fixed_ip, vrf_name)

    def _remove_floating_ip(self, ri, ext_gw_port, floating_ip, fixed_ip):
        vrf_name = self._get_vrf_name(ri)
        out_itfc_name = self._get_interface_name_from_hosting_port(ext_gw_port)
        # first remove NAT from outer interface
        self._remove_interface_nat(out_itfc_name, 'outside')
        # clear the NAT translation table
        self._remove_dyn_nat_translations()
        # remove the floating ip
        self._do_remove_floating_ip(floating_ip, fixed_ip, vrf_name)
        # enable NAT on outer interface
        self._add_interface_nat(out_itfc_name, 'outside')

    def _update_routing_table(self, ri, action, route):
        vrf_name = self._get_vrf_name(ri)
        destination_net = netaddr.IPNetwork(route['destination'])
        dest = destination_net.network
        dest_mask = destination_net.netmask
        next_hop = route['nexthop']
        if action is 'replace':
            self._add_static_route(dest, dest_mask, next_hop, vrf_name)
        elif action is 'delete':
            self._remove_static_route(dest, dest_mask, next_hop, vrf_name)
        else:
            LOG.error('Unknown route command %s', action)

    def _create_vrf(self, ri):
        vrf_name = self._get_vrf_name(ri)
        self._do_create_vrf(vrf_name)

    def _remove_vrf(self, ri):
        vrf_name = self._get_vrf_name(ri)
        self._do_remove_vrf(vrf_name)

    def _get_vrf_name(self, ri):
        return ri.router_name()[:self.DEV_NAME_LEN]

    def _get_connection(self):
        """Make SSH connection to the IOS XE device.

        The external ncclient library is used for creating this connection.
        This method keeps state of any existing connections and reuses them if
        already connected. Also interfaces (except management) are typically
        disabled by default when it is booted. So if connecting for the first
        time, driver will enable all other interfaces and keep that status in
        the `_itfcs_enabled` flag.
        """
        try:
            if self._ncc_connection and self._ncc_connection.connected:
                return self._ncc_connection
            else:
                # ncclient needs 'name' to be 'csr' in order to communicate
                # with the device in the correct way.
                self._ncc_connection = manager.connect(
                    host=self._host_ip, port=self._host_ssh_port,
                    username=self._username, password=self._password,
                    device_params={'name': "csr"}, timeout=self._timeout)
                if not self._itfcs_enabled:
                    self._itfcs_enabled = self._enable_itfcs(
                        self._ncc_connection)
            return self._ncc_connection
        except Exception as e:
            conn_params = {'host': self._host_ip, 'port': self._host_ssh_port,
                           'user': self._username,
                           'timeout': self._timeout, 'reason': e.message}
            raise cfg_exc.ConnectionException(**conn_params)

    def _get_interface_name_from_hosting_port(self, port):
        vlan = self._get_interface_vlan_from_hosting_port(port)
        return 'GigabitEthernet.%s' % vlan

    @staticmethod
    def _get_interface_vlan_from_hosting_port(port):
        return port['hosting_info']['segmentation_id']

    def _get_interfaces(self):
        """Get a list of interfaces on this hosting device.

        :return: List of the interfaces
        """
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        itfcs_raw = parse.find_lines("^interface GigabitEthernet")
        itfcs = [raw_if.strip().split(' ')[1] for raw_if in itfcs_raw]
        LOG.debug("Interfaces on hosting device: %s", itfcs)
        return itfcs

    def _get_interface_ip(self, interface_name):
        """Get the ip address for an interface.

        :param interface_name: interface_name as a string
        :return: ip address of interface as a string
        """
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        children = parse.find_children("^interface %s" % interface_name)
        for line in children:
            if 'ip address' in line:
                ip_address = line.strip().split(' ')[2]
                LOG.debug("IP Address:%s", ip_address)
                return ip_address
        LOG.warning("Cannot find interface: %s", interface_name)
        return None

    def _interface_exists(self, interface):
        """Check whether interface exists."""
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        itfcs_raw = parse.find_lines("^interface " + interface)
        return len(itfcs_raw) > 0

    def _enable_itfcs(self, conn):
        """Enable the interfaces of a IOS XE device.

        :param conn: Connection object
        :return: True or False
        """
        return True

    def _get_vrfs(self):
        """Get the current VRFs configured in the device.

        :return: A list of vrf names as string
        """
        vrfs = []
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        vrfs_raw = parse.find_lines("^vrf definition")
        for line in vrfs_raw:
            #  raw format ['ip vrf <vrf-name>',....]
            vrf_name = line.strip().split(' ')[2]
            vrfs.append(vrf_name)
        LOG.info("VRFs:%s", vrfs)
        return vrfs

    def _get_capabilities(self):
        """Get the servers NETCONF capabilities.

        :return: List of server capabilities.
        """
        conn = self._get_connection()
        capabilities = []
        for c in conn.server_capabilities:
            capabilities.append(c)
        LOG.debug("Server capabilities: %s", capabilities)
        return capabilities

    def _get_running_config(self, split=True):
        """Get the IOS XE device's current running config.

        :return: Current IOS running config as multiline string
        """
        conn = self._get_connection()
        config = conn.get_config(source="running")
        if config:
            root = ET.fromstring(config._raw)
            running_config = root[0][0]
            if split is True:
                rgx = re.compile("\r*\n+")
                ioscfg = rgx.split(running_config.text)
            else:
                ioscfg = running_config.text
            return ioscfg

    def _check_acl(self, acl_no, network, netmask):
        """Check a ACL config exists in the running config.

        :param acl_no: access control list (ACL) number
        :param network: network which this ACL permits
        :param netmask: netmask of the network
        :return:
        """
        exp_cfg_lines = ['ip access-list standard ' + str(acl_no),
                         ' permit ' + str(network) + ' ' + str(netmask)]
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        acls_raw = parse.find_children(exp_cfg_lines[0])
        if acls_raw:
            if exp_cfg_lines[1] in acls_raw:
                return True
            LOG.error("Mismatch in ACL configuration for %s", acl_no)
            return False
        LOG.debug("%s is not present in config", acl_no)
        return False

    def _cfg_exists(self, cfg_str):
        """Check a partial config string exists in the running config.

        :param cfg_str: config string to check
        :return : True or False
        """
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        cfg_raw = parse.find_lines("^" + cfg_str)
        LOG.debug("_cfg_exists(): Found lines %s", cfg_raw)
        return len(cfg_raw) > 0

    def _set_interface(self, name, ip_address, mask):
        conf_str = snippets.SET_INTC % (name, ip_address, mask)
        self._edit_running_config(conf_str, 'SET_INTC')

    def _do_create_vrf(self, vrf_name):
        conf_str = snippets.CREATE_VRF % vrf_name
        self._edit_running_config(conf_str, 'CREATE_VRF')

    def _do_remove_vrf(self, vrf_name):
        if vrf_name in self._get_vrfs():
            conf_str = snippets.REMOVE_VRF % vrf_name
            self._edit_running_config(conf_str, 'REMOVE_VRF')

    def _do_create_sub_interface(self, sub_interface, vlan_id, vrf_name, ip,
                                 mask):
        if vrf_name not in self._get_vrfs():
            LOG.error("VRF %s not present", vrf_name)
        conf_str = snippets.CREATE_SUBINTERFACE % (sub_interface, vlan_id,
                                                   vrf_name, ip, mask)
        self._edit_running_config(conf_str, 'CREATE_SUBINTERFACE')

    def _do_remove_sub_interface(self, sub_interface):
        # optional: verify this is the correct sub_interface
        if self._interface_exists(sub_interface):
            conf_str = snippets.REMOVE_SUBINTERFACE % sub_interface
            self._edit_running_config(conf_str, 'REMOVE_SUBINTERFACE')

    def _do_add_ha_hsrp(self, sub_interface, vrf_name, priority, group, ip):
        if vrf_name not in self._get_vrfs():
            LOG.error("VRF %s not present", vrf_name)
        conf_str = snippets.SET_INTC_HSRP % (sub_interface, vrf_name, group,
                                             priority, group, ip)
        action = "SET_INTC_HSRP (Group: %s, Priority: % s)" % (group, priority)
        self._edit_running_config(conf_str, action)

    def _do_remove_ha_hsrp(self, sub_interface, group):
        conf_str = snippets.REMOVE_INTC_HSRP % (sub_interface, group)
        action = ("REMOVE_INTC_HSRP (subinterface:%s, Group:%s)"
                  % (sub_interface, group))
        self._edit_running_config(conf_str, action)

    def _get_interface_cfg(self, interface):
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        return parse.find_children('interface ' + interface)

    def _nat_rules_for_internet_access(self, acl_no, network, netmask,
                                       inner_itfc, outer_itfc, vrf_name):
        """Configure the NAT rules for an internal network.

        :param acl_no: ACL number of the internal network.
        :param network: internal network
        :param netmask: netmask of the internal network.
        :param inner_itfc: (name of) interface connected to the internal
        network
        :param outer_itfc: (name of) interface connected to the external
        network
        :param vrf_name: VRF corresponding to this virtual router
        :return: True if configuration succeeded
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        IOSXEConfigException
        """
        pass

    def _add_interface_nat(self, itfc_name, itfc_type):
        conf_str = snippets.SET_NAT % (itfc_name, itfc_type)
        self._edit_running_config(conf_str, 'SET_NAT_' + itfc_type)

    def _remove_interface_nat(self, itfc_name, itfc_type):
        conf_str = snippets.REMOVE_NAT % (itfc_name, itfc_type)
        self._edit_running_config(conf_str, 'SET_NAT_' + itfc_type)

    def _remove_dyn_nat_rule(self, acl_no, outer_itfc_name, vrf_name):
        conf_str = snippets.SNAT_CFG % (acl_no, outer_itfc_name, vrf_name)
        if self._cfg_exists(conf_str):
            conf_str = snippets.REMOVE_DYN_SRC_TRL_INTFC % (
                acl_no, outer_itfc_name, vrf_name)
            self._edit_running_config(conf_str, 'REMOVE_DYN_SRC_TRL_INTFC')
        conf_str = snippets.REMOVE_ACL % acl_no
        self._edit_running_config(conf_str, 'REMOVE_ACL')

    def _remove_dyn_nat_translations(self):
        conf_str = snippets.CLEAR_DYN_NAT_TRANS
        self._edit_running_config(conf_str, 'CLEAR_DYN_NAT_TRANS')

    def _do_add_floating_ip(self, floating_ip, fixed_ip, vrf):
        conf_str = snippets.SET_STATIC_SRC_TRL % (fixed_ip, floating_ip, vrf)
        self._edit_running_config(conf_str, 'SET_STATIC_SRC_TRL')

    def _do_remove_floating_ip(self, floating_ip, fixed_ip, vrf):
        conf_str = snippets.REMOVE_STATIC_SRC_TRL % (
            fixed_ip, floating_ip, vrf)
        self._edit_running_config(conf_str, 'REMOVE_STATIC_SRC_TRL')

    def _get_floating_ip_cfg(self):
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        res = parse.find_lines('ip nat inside source static')
        return res

    def _add_static_route(self, dest, dest_mask, next_hop, vrf):
        conf_str = snippets.SET_IP_ROUTE % (vrf, dest, dest_mask, next_hop)
        self._edit_running_config(conf_str, 'SET_IP_ROUTE')

    def _remove_static_route(self, dest, dest_mask, next_hop, vrf):
        conf_str = snippets.REMOVE_IP_ROUTE % (vrf, dest, dest_mask, next_hop)
        self._edit_running_config(conf_str, 'REMOVE_IP_ROUTE')

    def _get_static_route_cfg(self):
        ios_cfg = self._get_running_config()
        parse = HTParser(ios_cfg)
        return parse.find_lines('ip route')

    def caller_name(self, skip=2):
        """
        Get a name of a caller in the format module.class.method

       `skip` specifies how many levels of stack to skip while getting caller
       name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

       An empty string is returned if skipped levels exceed stack height
       """
        stack = inspect.stack()
        start = 0 + skip
        if len(stack) < start + 1:
            return ''
        parentframe = stack[start][0]

        name = []
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        # TODO(asr1kteam): consider using __main__
        if module:
            name.append(module.__name__)
        # detect classname
        if 'self' in parentframe.f_locals:
            # I don't know any way to detect call from the object method
            # XXX: there seems to be no way to detect static method call,
            # it will be just a function call
            name.append(parentframe.f_locals['self'].__class__.__name__)
        codename = parentframe.f_code.co_name
        if codename != '<module>':  # top level usually
            name.append(codename)  # function or a method
        del parentframe
        return ".".join(name)

    # [ OR ]
    # curframe = inspect.currentframe()
    # calframe = inspect.getouterframes(curframe, 2)
    # return calframe[1][3]

    def _edit_running_config(self, conf_str, snippet):
        conn = self._get_connection()
        LOG.info("Config generated for [%(device)s] %(snip)s is:%(conf)s "
                 "caller:%(caller)s",
                 {'device': self.hosting_device['id'],
                  'snip': snippet,
                  'conf': conf_str,
                  'caller': self.caller_name()})
        try:
            rpc_obj = conn.edit_config(target='running', config=conf_str)
            self._check_response(rpc_obj, snippet, conf_str=conf_str)
        except Exception as e:
            # Here we catch all exceptions caused by REMOVE_/DELETE_ configs
            # to avoid config agent to get stuck once it hits this condition.
            # This is needed since the current ncclient version (0.4.2)
            # generates an exception when an attempt to configure the device
            # fails by the device (ASR1K router) but it doesn't provide any
            # details about the error message that the device reported.
            # With ncclient 0.4.4 version and onwards the exception returns
            # also the proper error. Hence this code can be changed when the
            # ncclient version is increased.
            if re.search(r"REMOVE_|DELETE_", snippet):
                LOG.warning("Pass exception for %s", snippet)
                pass
            elif isinstance(e, ncclient.operations.rpc.RPCError):
                e_tag = e.tag
                e_type = e.type
                params = {'snippet': snippet, 'type': e_type, 'tag': e_tag,
                          'dev_id': self.hosting_device['id'],
                          'ip': self._host_ip, 'confstr': conf_str}
                raise cfg_exc.IOSXEConfigException(**params)

    def _check_response(self, rpc_obj, snippet_name, conf_str=None):
        """This function checks the rpc response object for status.

        This function takes as input the response rpc_obj and the snippet name
        that was executed. It parses it to see, if the last edit operation was
        a success or not.
            <?xml version="1.0" encoding="UTF-8"?>
            <rpc-reply message-id="urn:uuid:81bf8082-....-b69a-000c29e1b85c"
                       xmlns="urn:ietf:params:netconf:base:1.0">
                <ok />
            </rpc-reply>
        In case of error, IOS XE device sends a response as follows.
        We take the error type and tag.
            <?xml version="1.0" encoding="UTF-8"?>
            <rpc-reply message-id="urn:uuid:81bf8082-....-b69a-000c29e1b85c"
            xmlns="urn:ietf:params:netconf:base:1.0">
                <rpc-error>
                    <error-type>protocol</error-type>
                    <error-tag>operation-failed</error-tag>
                    <error-severity>error</error-severity>
                </rpc-error>
            </rpc-reply>
        :return: True if the config operation completed successfully
        :raises: networking_cisco.plugins.cisco.cfg_agent.cfg_exceptions.
        IOSXEConfigException
        """
        LOG.debug("RPCReply for %(snippet_name)s is %(rpc_obj)s",
                  {'snippet_name': snippet_name, 'rpc_obj': rpc_obj.xml})
        xml_str = rpc_obj.xml
        if "<ok />" in xml_str:
            # LOG.debug("RPCReply for %s is OK", snippet_name)
            LOG.info("%s was successfully executed", snippet_name)
            return True
        # Not Ok, we throw a ConfigurationException
        e_type = rpc_obj._root[0][0].text
        e_tag = rpc_obj._root[0][1].text
        params = {'snippet': snippet_name, 'type': e_type, 'tag': e_tag,
                  'dev_id': self.hosting_device['id'],
                  'ip': self._host_ip, 'confstr': conf_str}
        raise cfg_exc.IOSXEConfigException(**params)



# Copyright 2017 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import mock

from neutron.tests import base

from networking_cisco.apps.saf.agent.topo_disc import pub_lldp_api as pub_lldp
from networking_cisco.apps.saf.common import dfa_sys_lib as sys_utils

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class LldpApiTest(base.BaseTestCase):
    """A test suite to exercise the public LldpApi class. """

    def setUp(self):
        """Setup for LldpApiTest. """
        super(LldpApiTest, self).setUp()
        self.root_helper = 'sudo'
        self.port_name = 'eth2'
        self.chassis_id_tlv = (
            "Chassis ID TLV\n\n\tMAC: e0:0e:da:cc:ac:1d\n\n")
        self.portid_tlv = (
            "Port ID TLV\n\n\tIfname: Ethernet1/2\n\t"
            "MAC: e0:0e:da:cc:ac:1e\n\tLocal: eth2\n")
        self.ttl_tlv = "Time to Live TLV\n\n\t120\n\n"
        self.port_descr_tlv = "Port Description TLV\n\tEthernet1/2\n"
        self.system_name_tlv = "System Name TLV\n\tN9k-L4\n"
        self.system_desc_tlv = (
            "System Description TLV\n\tCisco Nexus Operating System (NX-OS) "
            "Software 7.0(3)I5(2)TAC support: http://www.cisco.com/tac "
            "Copyright (c) 2002-2016, Cisco Systems, Inc. All rights "
            "reserved.\n")
        self.system_capab_tlv = (
            "System Capabilities TLV\n\t"
            "System capabilities:  Bridge, Router\n\t"
            "Enabled capabilities: Bridge, Router\n\t")
        self.management_address_tlv = (
            "Management Address TLV\n\tIPv4: 11.0.100.74\n\t"
            "Ifindex: 83886080\n\t")
        self.cisco_power_tlv = (
            "Cisco 4-wire Power-via-MDI TLV\n\t4-Pair PoE supported\n\t"
            "Spare pair Detection/Classification not required\n\t"
            "PD Spare pair Desired State: Disabled\n\t"
            "PSE Spare pair Operational State: Disabled\n")
        self.port_vlan_id_tlv = "Port VLAN ID TLV\n\tPVID: 1\n"
        self.mgmt_address_tlv = (
            "Management Address TLV\n\tIPv6: 2002:11:0:100::74\n\t"
            "Ifindex: 83886080\n")
        self.evb_cfg_tlv = (
            "EVB Configuration TLV\n\tbridge:bgid(0x4)\n\tstation:(00)\n\t"
            "retries:3 rte:14\n\tmode:bridge r/l:0 rwd:20\n\tr/l:0 rka:22\n\t")
        self.end_tlv = "End of LLDPDU TLV"
        self.lldp_tlv_correct_out = ''.join(
            (self.chassis_id_tlv, self.portid_tlv, self.ttl_tlv,
             self.port_descr_tlv, self.system_name_tlv, self.system_desc_tlv,
             self.system_capab_tlv, self.management_address_tlv,
             self.cisco_power_tlv, self.port_vlan_id_tlv,
             self.mgmt_address_tlv, self.evb_cfg_tlv, self.end_tlv))
        self.execute = mock.patch.object(
            sys_utils, "execute", spec=sys_utils.execute).start()
        self.pub_lldp = pub_lldp.LldpApi(self.root_helper)

    def test_enable_lldp_ncb_correct_reply(self):
        """Test for enable_lldp function for correct return value. """
        self.execute.return_value = "adminstatus=rxtx"
        ret = self.pub_lldp.enable_lldp(self.port_name)
        self.execute.assert_called_with(
            ["lldptool", "-L", "-i", self.port_name, "-g", "ncb",
             "adminStatus=rxtx"], root_helper=self.root_helper)
        self.assertTrue(ret)

    def test_enable_lldp_ncb_incorrect_reply(self):
        """Test for enable_lldp function for incorrect return value. """
        self.execute.return_value = ""
        ret = self.pub_lldp.enable_lldp(self.port_name)
        self.execute.assert_called_with(
            ["lldptool", "-L", "-i", self.port_name, "-g", "ncb",
             "adminStatus=rxtx"], root_helper=self.root_helper)
        self.assertFalse(ret)

    def test_enable_lldp_nb(self):
        """Test for enable_lldp function for NB DMAC case. """
        self.execute.return_value = "adminstatus=rxtx"
        ret = self.pub_lldp.enable_lldp(self.port_name,
                                        is_ncb=False, is_nb=True)
        self.execute.assert_called_with(
            ["lldptool", "-L", "-i", self.port_name, "-g", "nb",
             "adminStatus=rxtx"], root_helper=self.root_helper)
        self.assertTrue(ret)

    def test_enable_lldp_invalid_case(self):
        """Test for enable_lldp function for neither NCB or NB DMAC case. """
        self.execute.return_value = "adminstatus=rxtx"
        ret = self.pub_lldp.enable_lldp(self.port_name, is_ncb=False)
        self.execute.assert_not_called()
        self.assertFalse(ret)

    def test_get_lldp_tlv_ncb(self):
        """Test for get_lldp_tlv for ncb DMAC case. """
        self.execute.return_value = 'some TLV'
        ret = self.pub_lldp.get_lldp_tlv(self.port_name, is_ncb=True)
        self.execute.assert_called_with(
            ["lldptool", "get-tlv", "-n", "-i", self.port_name, "-g",
             "ncb"], root_helper=self.root_helper)
        self.assertEqual(self.execute.return_value, ret)

    def test_get_lldp_tlv_nb(self):
        """Test for get_lldp_tlv for nb DMAC case. """
        self.execute.return_value = 'some TLV'
        ret = self.pub_lldp.get_lldp_tlv(self.port_name, is_ncb=False,
                                         is_nb=True)
        self.execute.assert_called_with(
            ["lldptool", "get-tlv", "-n", "-i", self.port_name, "-g",
             "nb"], root_helper=self.root_helper)
        self.assertEqual(self.execute.return_value, ret)

    def test_common_tlv_format_none_case(self):
        """Test _check_common_tlv_format when TLV data is None. """
        ret, tlv_parsed = self.pub_lldp._check_common_tlv_format(
            None, None, None)
        self.assertFalse(ret)
        self.assertIsNone(tlv_parsed)

    def test_common_tlv_format_no_tlv_case(self):
        """Test _check_common_tlv_format when specific TLV is not present. """
        invalid_tlv = self.lldp_tlv_correct_out.replace(self.evb_cfg_tlv, '')
        ret, tlv_parsed = self.pub_lldp._check_common_tlv_format(
            invalid_tlv, "mode:", "EVB Configuration TLV")
        self.assertFalse(ret)
        self.assertIsNone(tlv_parsed)

    def test_common_tlv_format_no_tlv_data_case(self):
        """Test _check_common_tlv_format when TLV data is not present. """
        invalid_tlv = self.lldp_tlv_correct_out.replace(self.evb_cfg_tlv, '')
        invalid_tlv = invalid_tlv + "EVB Configuration TLV\n\t"
        ret, tlv_parsed = self.pub_lldp._check_common_tlv_format(
            invalid_tlv, "mode:", "EVB Configuration TLV")
        self.assertFalse(ret)
        self.assertIsNone(tlv_parsed)

    def test_common_tlv_format_no_tlv_data_pattern_case(self):
        """Test _check_common_tlv_format when TLV pattern is not present. """
        invalid_tlv = self.lldp_tlv_correct_out.replace(
            "mode:bridge r/l:0 rwd:20", '')
        ret, tlv_parsed = self.pub_lldp._check_common_tlv_format(
            invalid_tlv, "mode:", "EVB Configuration TLV")
        self.assertFalse(ret)
        self.assertIsNone(tlv_parsed)

    def test_remote_evb_cfgd(self):
        """Test the case when remove EVB TLV is present.

        False case is tested in the test_common.. cases.
        """
        ret = self.pub_lldp.get_remote_evb_cfgd(self.lldp_tlv_correct_out)
        self.assertTrue(ret)

    def test_remote_evb_mode(self):
        """Test the get_remote_evb_mode function. """
        ret = self.pub_lldp.get_remote_evb_mode(self.lldp_tlv_correct_out)
        self.assertEqual('bridge', ret)

    def test_remote_evb_mode_incorrect(self):
        """Test the get_remote_evb_mode function for incorrect TLV. """
        invalid_tlv = self.lldp_tlv_correct_out.replace(
            "mode:bridge r/l:0 rwd:20", '')
        ret = self.pub_lldp.get_remote_evb_mode(invalid_tlv)
        self.assertIsNone(ret)

    def test_remote_mgmt_addr(self):
        """Test the get_remote_mgmt_addr function. """
        ret = self.pub_lldp.get_remote_mgmt_addr(self.lldp_tlv_correct_out)
        self.assertEqual('IPv4:11.0.100.74', ret)

    def test_remote_system_desc(self):
        """Test the get_remote_system_desc function. """
        ret = self.pub_lldp.get_remote_sys_desc(self.lldp_tlv_correct_out)
        match_string = self.system_desc_tlv.replace(
            "System Description TLV", '').replace('\n', '').replace('\t', '')
        self.assertEqual(match_string, ret)

    def test_remote_system_name(self):
        """Test the get_remote_system_name function. """
        ret = self.pub_lldp.get_remote_sys_name(self.lldp_tlv_correct_out)
        self.assertEqual('N9k-L4', ret)

    def test_remote_port(self):
        """Test the get_remote_port function. """
        ret = self.pub_lldp.get_remote_port(self.lldp_tlv_correct_out)
        self.assertEqual('Ethernet1/2', ret)

    def test_remote_chassis_id_mac(self):
        """Test the get_remote_chassis_id_mac function. """
        ret = self.pub_lldp.get_remote_chassis_id_mac(
            self.lldp_tlv_correct_out)
        self.assertEqual('e0:0e:da:cc:ac:1d', ret)

    def test_remote_port_id_mac(self):
        """Test the get_remote_port_id_mac function. """
        ret = self.pub_lldp.get_remote_port_id_mac(self.lldp_tlv_correct_out)
        self.assertEqual('e0:0e:da:cc:ac:1e', ret)

    def test_remote_port_id_local(self):
        """Test the get_remote_port_id_local function. """
        ret = self.pub_lldp.get_remote_port_id_local(self.lldp_tlv_correct_out)
        self.assertEqual('eth2', ret)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
import subprocess
import sys
import time

from networking_cisco.apps.saf.common import dfa_logger as logging

LOG = logging.getLogger(__name__)

uplink_file_path = '/tmp/uplink'
detect_uplink_file_path = '/tmp/uplink_detected'


def run_cmd_line(cmd_str, stderr=None, shell=True,
                 echo_cmd=False, check_result=False):
    if echo_cmd:
        LOG.debug(cmd_str)
    if shell:
        cmd_args = cmd_str
    else:
        cmd_args = cmd_str.split()
    output = None
    returncode = 0
    try:
        output = subprocess.check_output(cmd_args, shell=shell, stderr=stderr)
    except subprocess.CalledProcessError as e:
        if check_result:
            LOG.debug(e)
            sys.exit(e.returncode)
        else:
            returncode = e.returncode
    return output, returncode


def read_file(file_name):
    file_content = None
    if os.path.isfile(file_name):
        filep = open(file_name, "r")
        file_content = filep.read()
        filep.close()
        file_content = file_content.replace("\n", "")
    return file_content


def find_uplink():
    intf_cmd_list = ("ip link |grep 'state UP' | awk '{print $2}' "
                     "| sed 's/://'|grep ^[epb]")
    intf_net_addr = "ifconfig %s | grep 'inet addr'"
    en_rxtx = ('sudo /usr/sbin/lldptool -i %s -g "ncb" -L adminStatus=rxtx')
    dis_rxtx = ('sudo /usr/sbin/lldptool -i %s -g "ncb" -L '
                'adminStatus=disabled')
    mod_brdg = ('sudo /usr/sbin/lldptool -i %s -g "ncb" -t -n -V evb | '
                'grep "mode:bridge"')

    intf_list, returncode = run_cmd_line(intf_cmd_list)
    for intf in intf_list.split():
        intf_out, retcode = run_cmd_line(intf_net_addr % intf)
        if intf_out is None:
            out, ret = run_cmd_line(en_rxtx % intf)
            time.sleep(40)
            out, ret = run_cmd_line(mod_brdg % intf)
            run_cmd_line(dis_rxtx % intf)
            if out:
                return intf


def detect_uplink_non_auto(input_string):
    file_str = "normal"
    if input_string is None:
        file_str = read_file(uplink_file_path)
    return file_str


def detect_uplink_auto(input_string):
    if input_string is None:
        return_str = find_uplink()
    else:
        cmd_str = ('sudo /usr/sbin/lldptool -i %s -g "ncb" -t -n -V evb | '
                   'grep "mode:bridge"') % input_string
        (output, returncode) = run_cmd_line(cmd_str,
                                            check_result=False)
        if returncode == 0:
            return_str = "normal"
        else:
            return_str = "down"

    LOG.debug('return_str=%s', return_str)
    return return_str


def detect_uplink(input_string=None):
    auto_detect = False
    if os.path.isfile(uplink_file_path):
        detected_uplink = detect_uplink_non_auto(input_string)
    else:
        detected_uplink = detect_uplink_auto(input_string)
        auto_detect = True
    log_str = "auto detect = %s, input string %s, detected uplink is %s." % (
        auto_detect, input_string, detected_uplink)
    LOG.debug(log_str)
    return detected_uplink



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import sqlalchemy as sa

from networking_cisco import backwards_compatibility as bc


class NexusPortBinding(bc.model_base.BASEV2):
    """Represents a binding of VM's to nexus ports."""

    __tablename__ = "cisco_ml2_nexusport_bindings"

    binding_id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    port_id = sa.Column(sa.String(255))
    vlan_id = sa.Column(sa.Integer, nullable=False)
    vni = sa.Column(sa.Integer)
    switch_ip = sa.Column(sa.String(255))
    instance_id = sa.Column(sa.String(255))
    channel_group = sa.Column(sa.Integer, default=0)
    is_native = sa.Column(sa.Boolean(), nullable=False, default=False,
                          server_default=sa.sql.false())

    def __repr__(self):
        """Just the binding, without the id key."""
        return ("<NexusPortBinding(%s,%s,%s,%s,%s,%s,%s)>" %
                (self.port_id, self.vlan_id, self.vni, self.switch_ip,
                 self.instance_id,
                 self.channel_group,
                 'True' if self.is_native else 'False'))

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        """Compare only the binding, without the id key."""
        return (
            self.port_id == other.port_id and
            self.vlan_id == other.vlan_id and
            self.vni == other.vni and
            self.switch_ip == other.switch_ip and
            self.instance_id == other.instance_id and
            self.channel_group == other.channel_group and
            self.is_native_vlan == other.is_native_vlan
        )


class NexusNVEBinding(bc.model_base.BASEV2):
    """Represents Network Virtualization Endpoint configuration."""

    __tablename__ = "cisco_ml2_nexus_nve"

    vni = sa.Column(sa.Integer, primary_key=True, nullable=False)
    device_id = sa.Column(sa.String(255), primary_key=True)
    switch_ip = sa.Column(sa.String(255), primary_key=True)
    mcast_group = sa.Column(sa.String(255))

    def __repr__(self):
        return ("<NexusNVEBinding(%s,%s,%s,%s)>" %
                (self.vni, self.switch_ip, self.device_id, self.mcast_group))


class NexusVxlanAllocation(bc.model_base.BASEV2):

    __tablename__ = 'ml2_nexus_vxlan_allocations'

    vxlan_vni = sa.Column(sa.Integer, nullable=False, primary_key=True,
                          autoincrement=False)
    allocated = sa.Column(sa.Boolean, nullable=False, default=False,
                          server_default=sa.sql.false())


class NexusMcastGroup(bc.model_base.BASEV2, bc.model_base.HasId):

    __tablename__ = 'ml2_nexus_vxlan_mcast_groups'

    mcast_group = sa.Column(sa.String(64), nullable=False)
    associated_vni = sa.Column(sa.Integer,
                               sa.ForeignKey(
                                   'ml2_nexus_vxlan_allocations.vxlan_vni',
                                   ondelete="CASCADE"),
                               nullable=False)


class NexusHostMapping(bc.model_base.BASEV2):

    """Nexus Host to interface Mapping."""

    __tablename__ = 'cisco_ml2_nexus_host_interface_mapping'

    host_id = sa.Column(sa.String(255), nullable=False, primary_key=True,
                        index=True)
    switch_ip = sa.Column(sa.String(255), nullable=False, primary_key=True,
                          index=True)
    if_id = sa.Column(sa.String(255), nullable=False, primary_key=True)
    ch_grp = sa.Column(sa.Integer(), nullable=False)
    is_static = sa.Column(sa.Boolean(), nullable=False)


class NexusVPCAlloc(bc.model_base.BASEV2):
    """Nexus Port Channel Allocation."""

    __tablename__ = 'cisco_ml2_nexus_vpc_alloc'

    switch_ip = sa.Column(sa.String(64), nullable=False,
                          primary_key=True)
    vpc_id = sa.Column(sa.Integer, nullable=False, primary_key=True,
                       autoincrement=False)

    # learned indicates this ch_grp was predefined on Nexus Switch
    # as opposed to ML2 Nexus driver creating the port channel #
    learned = sa.Column(sa.Boolean, nullable=False)

    # active indicates this ch_grp is in use
    active = sa.Column(sa.Boolean, nullable=False)


class NexusProviderNetwork(bc.model_base.BASEV2):

    __tablename__ = 'cisco_ml2_nexus_provider_networks'

    network_id = sa.Column(sa.String(36), nullable=False, primary_key=True)
    vlan_id = sa.Column(sa.Integer, nullable=False, index=True)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""ha_for_routing_service_in_cisco_devices

Revision ID: 1e9e22602685
Revises: 53f08de0523f
Create Date: 2015-09-28 09:33:27.294138

"""

# revision identifiers, used by Alembic.
revision = '1e9e22602685'
down_revision = '2921fe565328'

from alembic import op
import sqlalchemy as sa

from networking_cisco.plugins.cisco.extensions import ha


ha_states = sa.Enum('ACTIVE', 'STANDBY', name='ha_states')


def upgrade():
    op.create_table('cisco_router_ha_settings',
        sa.Column('router_id', sa.String(36), nullable=True),
        sa.Column('ha_type', sa.String(255), nullable=True),
        sa.Column('redundancy_level', sa.Integer,
                  server_default=str(ha.MIN_REDUNDANCY_LEVEL)),
        sa.Column('priority', sa.Integer, nullable=True),
        sa.Column('probe_connectivity', sa.Boolean, nullable=True),
        sa.Column('probe_target', sa.String(64), nullable=True),
        sa.Column('probe_interval', sa.Integer, nullable=True),
        sa.Column('state', ha_states, server_default='ACTIVE'),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('router_id')
    )
    op.create_table('cisco_router_ha_groups',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('ha_type', sa.String(255), nullable=True),
        sa.Column('group_identity', sa.String(255), nullable=True),
        sa.Column('ha_port_id', sa.String(36), nullable=False),
        sa.Column('extra_port_id', sa.String(36), nullable=True),
        sa.Column('subnet_id', sa.String(36), nullable=True),
        sa.Column('user_router_id', sa.String(36), nullable=True),
        sa.Column('timers_config', sa.String(255), nullable=True),
        sa.Column('tracking_config', sa.String(255), nullable=True),
        sa.Column('other_config', sa.String(255), nullable=True),
        sa.ForeignKeyConstraint(['ha_port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['extra_port_id'], ['ports.id'],
                                ondelete='SET NULL'),
        sa.ForeignKeyConstraint(['subnet_id'], ['subnets.id']),
        sa.ForeignKeyConstraint(['user_router_id'], ['routers.id']),
        sa.PrimaryKeyConstraint('ha_port_id')
    )
    op.create_table('cisco_router_redundancy_bindings',
        sa.Column('redundancy_router_id', sa.String(36)),
        sa.Column('priority', sa.Integer),
        sa.Column('state', ha_states, server_default='STANDBY'),
        sa.Column('user_router_id', sa.String(36)),
        sa.ForeignKeyConstraint(['redundancy_router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_router_id'], ['routers.id']),
        sa.PrimaryKeyConstraint('redundancy_router_id')
    )



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from networking_cisco.plugins.cisco.cpnr import dhcpopts


class TestDhcpopts(base.BaseTestCase):
    def test_format_for_classless_static_routes_options(self):
        expected = '20:a9:fe:a9:fe:0a:0a:00:02'
        value = dhcpopts.format_for_options('classless-static-routes',
                        '32.169.254.169.254 10.10.0.2')
        self.assertEqual(expected, value)

    def test_format_for_ip_forwarding_options(self):
        expected = '01'
        flag = True
        value = dhcpopts.format_for_options('ip-forwarding', flag)
        self.assertEqual(expected, value)

    def test_format_for_dhcp_renewal_time_options(self):
        expected = '00:00:01:2c'
        value = dhcpopts.format_for_options('dhcp-renewal-time', 300)
        self.assertEqual(expected, value)

    def test_format_for_domain_name_options(self):
        expected = '65:78:61:6d:70:6c:65:2e:63:6f:6d:2e'
        value = dhcpopts.format_for_options('domain-name', 'example.com.')
        self.assertEqual(expected, value)

    @mock.patch('networking_cisco.plugins.cisco.cpnr.dhcpopts._format_value')
    def test_format_for_options_exception(self, mock_format):
        expected = ''
        mock_format.get.side_effect = Exception
        value = dhcpopts.format_for_options('subnet-mask', '10.10.0.1')
        self.assertEqual(expected, value)

    @mock.patch('networking_cisco.plugins.cisco.cpnr.dhcpopts._format_value')
    def test_format_for_options_unknown_option(self, mock_format):
        expected = None
        value = dhcpopts.format_for_options('someoption', 'somevalue')
        self.assertEqual(expected, value)

    def test_format_route_list_value(self):
        expected = '20a9fea9fe0a0a0002'
        value = dhcpopts._format_value('route-list',
                        '32.169.254.169.254 10.10.0.2')
        self.assertEqual(expected, value)

    def test_format_bool_value(self):
        expected = '01'
        flag = True
        value = dhcpopts._format_value('bool', flag)
        self.assertEqual(expected, value)

    def test_format_int32_value(self):
        expected = '0000012c'
        value = dhcpopts._format_value('int32', 300)
        self.assertEqual(expected, value)

    def test_format_string_value(self):
        expected = '6578616d706c652e636f6d2e'
        value = dhcpopts._format_value('string', 'example.com.')
        self.assertEqual(expected, value)

    def test_format_ip_value(self):
        expected = '0a0a0001'
        value = dhcpopts._format_value('ip', '10.10.0.1')
        self.assertEqual(expected, value)

    def test_format_comma_separated_value(self):
        expected = '0a0a00016578616d706c652e636f6d'
        value = dhcpopts._format_value('ip, string', '10.10.0.1 example.com')
        self.assertEqual(expected, value)

    def test_format_none_value(self):
        expected = ''
        value = dhcpopts._format_value('none', 'none')
        self.assertEqual(expected, value)

    def test_format_route_list_value_exception(self):
        expected = ''
        try:
            value = dhcpopts._format_value('route-list',
                        '32.169.254.169.254 ip-adress')
        except ValueError:
            value = ''
        self.assertEqual(expected, value)



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.server.services.firewall.native import (
    fw_constants as fw_const)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    dev_mgr_plug)

LOG = logging.getLogger(__name__)


# Not sure of the exact name. But, this implements a case when all requests
# goto first device until it exhausts
class MaxSched(object):

    """Max Sched class.

    This scheduler will return the first firewall until it reaches its quota.
    """

    def __init__(self, obj_dict):
        """Initialization. """
        self.num_res = len(obj_dict)
        self.obj_dict = obj_dict
        self.res = dict()
        cnt = 0
        for ip in self.obj_dict:
            obj_elem_dict = self.obj_dict.get(ip)
            drvr_obj = obj_elem_dict.get('drvr_obj')
            self.res[cnt] = {'mgmt_ip': ip,
                             'quota': drvr_obj.get_max_quota(),
                             'obj_dict': obj_elem_dict,
                             'used': 0,
                             'fw_id_lst': []}
            cnt += 1

    def allocate_fw_dev(self, fw_id):
        """Allocate firewall device.

        Allocate the first Firewall device which has resources available.
        """
        for cnt in self.res:
            used = self.res.get(cnt).get('used')
            if used < self.res.get(cnt).get('quota'):
                self.res[cnt]['used'] = used + 1
                self.res[cnt]['fw_id_lst'].append(fw_id)
                return self.res[cnt].get('obj_dict'), (
                    self.res[cnt].get('mgmt_ip'))
        return None, None

    def populate_fw_dev(self, fw_id, mgmt_ip, new):
        """Populate the class after a restart. """
        for cnt in self.res:
            used = self.res.get(cnt).get('used')
            if mgmt_ip == self.res[cnt].get('mgmt_ip'):
                if new:
                    self.res[cnt]['used'] = used + 1
                self.res[cnt]['fw_id_lst'].append(fw_id)
                return self.res[cnt].get('obj_dict'), (
                    self.res[cnt].get('mgmt_ip'))
        return None, None

    def get_fw_dev_map(self, fw_id):
        """Return the object dict and mgmt ip for a firewall. """
        for cnt in self.res:
            if fw_id in self.res.get(cnt).get('fw_id_lst'):
                return self.res[cnt].get('obj_dict'), (
                    self.res[cnt].get('mgmt_ip'))
        return None, None

    def deallocate_fw_dev(self, fw_id):
        """Release the firewall resource. """
        for cnt in self.res:
            if fw_id in self.res.get(cnt).get('fw_id_lst'):
                self.res[cnt]['used'] = self.res[cnt]['used'] - 1
                self.res.get(cnt).get('fw_id_lst').remove(fw_id)
                return


class DeviceMgr(object):

    """Device Manager. """

    def __init__(self, cfg):
        """Initialization. """
        self.drvr_obj = {}
        self.mgmt_ip_list = cfg.firewall.fw_mgmt_ip
        self.mgmt_ip_list = self.mgmt_ip_list.strip('[').rstrip(']').split(',')
        self.user_list = cfg.firewall.fw_username
        if self.user_list:
            self.user_list = self.user_list.strip('[').rstrip(']').split(',')
        self.pwd_list = cfg.firewall.fw_password
        if self.pwd_list:
            self.pwd_list = self.pwd_list.strip('[').rstrip(']').split(',')
        self.interface_in_list = cfg.firewall.fw_interface_in
        if self.interface_in_list:
            self.interface_in_list = self.interface_in_list.strip('[').\
                rstrip(']').split(',')
        self.interface_out_list = cfg.firewall.fw_interface_out
        if self.interface_out_list:
            self.interface_out_list = self.interface_out_list.strip('[').\
                rstrip(']').split(',')
        self.obj_dict = dict()
        cnt = 0
        dev = cfg.firewall.device
        # Modify enabler_conf.ini in source path for IP list
        for ip in self.mgmt_ip_list:
            ip = ip.strip()
            obj = dev_mgr_plug.DeviceMgr(cfg, dev)
            self.obj_dict[ip] = dict()
            self.obj_dict[ip]['drvr_obj'] = obj.get_drvr_obj()
            self.obj_dict[ip]['dev_name'] = cfg.firewall.device.split(',')[cnt]
            cnt = cnt + 1
        self.drvr_initialize(cfg)
        if cfg.firewall.sched_policy == fw_const.SCHED_POLICY:
            self.sched_obj = MaxSched(self.obj_dict)

    def populate_local_sch_cache(self, fw_dict):
        """Populate the local cache from FW DB after restart. """
        for fw_id in fw_dict:
            fw_data = fw_dict.get(fw_id)
            mgmt_ip = fw_data.get('fw_mgmt_ip')
            dev_status = fw_data.get('device_status')
            if dev_status == 'SUCCESS':
                new = True
            else:
                new = False
            if mgmt_ip is not None:
                drvr_dict, mgmt_ip = self.sched_obj.populate_fw_dev(fw_id,
                                                                    mgmt_ip,
                                                                    new)
                if drvr_dict is None or mgmt_ip is None:
                    LOG.info("Pop cache for FW sch: drvr_dict or mgmt_ip "
                             "is None")

    def drvr_initialize(self, cfg):
        """Initialize the driver routines. """
        cnt = 0
        for ip in self.obj_dict:
            cfg_dict = {}
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            cfg_dict['mgmt_ip_addr'] = ip
            if self.user_list is not None:
                cfg_dict['user'] = self.user_list[cnt]
            if self.pwd_list is not None:
                cfg_dict['pwd'] = self.pwd_list[cnt]
            if self.interface_in_list is not None:
                cfg_dict['interface_in'] = self.interface_in_list[cnt]
            if self.interface_out_list is not None:
                cfg_dict['interface_out'] = self.interface_out_list[cnt]
            drvr_obj.initialize(cfg_dict)
            cnt = cnt + 1

    def populate_event_que(self, que_obj):
        """Populates the event queue object.

        This is for sending router events to event handler.
        """
        for ip in self.obj_dict:
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            drvr_obj.populate_event_que(que_obj)

    def populate_dcnm_obj(self, dcnm_obj):
        """Populates the DCNM object. """
        for ip in self.obj_dict:
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            drvr_obj.populate_dcnm_obj(dcnm_obj)

    def is_device_virtual(self):
        """Returns if the device is physical or virtual. """
        for ip in self.obj_dict:
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            ret = drvr_obj.is_device_virtual()
            # No way to pin a device as of now, so return the first
            # TODO(padkrish)
            return ret

    def create_fw_device(self, tenant_id, fw_id, data):
        """Creates the Firewall. """
        drvr_dict, mgmt_ip = self.sched_obj.allocate_fw_dev(fw_id)
        if drvr_dict is not None and mgmt_ip is not None:
            self.update_fw_db_mgmt_ip(fw_id, mgmt_ip)
            ret = drvr_dict.get('drvr_obj').create_fw(tenant_id, data)
            if not ret:
                self.sched_obj.deallocate_fw_dev(fw_id)
            return ret
        else:
            return False

    def delete_fw_device(self, tenant_id, fw_id, data):
        """Deletes the Firewall. """
        drvr_dict, mgmt_ip = self.sched_obj.get_fw_dev_map(fw_id)
        ret = drvr_dict.get('drvr_obj').delete_fw(tenant_id, data)
        # FW DB gets deleted, so no need to remove the MGMT IP
        if ret:
            self.sched_obj.deallocate_fw_dev(fw_id)
        return ret

    def modify_fw_device(self, tenant_id, fw_id, data):
        """Modifies the firewall cfg. """
        drvr_dict, mgmt_ip = self.sched_obj.get_fw_dev_map(fw_id)
        return drvr_dict.get('drvr_obj').modify_fw(tenant_id, data)

    def network_create_notif(self, tenant_id, tenant_name, cidr):
        """Notification for Network create.

        Since FW ID not present, it's not possible to know which FW instance
        to call. So, calling everyone, each instance will figure out if it
        applies to them.
        """
        for ip in self.obj_dict:
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            ret = drvr_obj.network_create_notif(tenant_id, tenant_name, cidr)
            LOG.info("Driver with IP %(ip)s return %(ret)s",
                     {'ip': ip, 'ret': ret})

    def network_delete_notif(self, tenant_id, tenant_name, net_id):
        """Notification for Network delete.

        Since FW ID not present, it's not possible to know which FW instance
        to call. So, calling everyone, each instance will figure out if it
        applies to them.
        """
        for ip in self.obj_dict:
            drvr_obj = self.obj_dict.get(ip).get('drvr_obj')
            ret = drvr_obj.network_delete_notif(tenant_id, tenant_name,
                                                net_id)
            LOG.info("Driver with IP %(ip)s return %(ret)s for network "
                     "delete notification", {'ip': ip, 'ret': ret})



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd and not os.path.isdir("contributor/api"):
    print("On Read the Docs and autodoc python module docs aren't built. "
          "Building...")
    os.environ['READTHEDOCS'] = 'False'
    os.environ['JUST_BUILD_AUTO_DOC'] = 'True'
    subprocess.check_call(
        [sys.executable, 'setup.py', 'build_sphinx', '-b', 'dummy'],
        cwd=os.path.join(os.path.dirname(__file__), os.path.pardir,
                         os.path.pardir),
    )
    os.environ['JUST_BUILD_AUTO_DOC'] = 'False'
    os.environ['READTHEDOCS'] = 'True'

ignore_everything = os.environ.get('JUST_BUILD_AUTO_DOC', None) == 'True'
if ignore_everything:
    master_doc = 'dummy'
    exclude_patterns = ['index.rst', '*/*']
else:
    # The master toctree document.
    master_doc = 'index'
    exclude_patterns = ['dummy.rst', 'api/networking_cisco_tempest_plugin.*']

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'reno.sphinxext',
    'oslo_config.sphinxext',
    'oslo_config.sphinxconfiggen',
    #'sphinx.ext.intersphinx',
]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# The suffix of source filenames.
source_suffix = '.rst'

# General information about the project.
project = u'networking-cisco'
copyright = u'2017, Cisco Systems, Inc'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
html_static_path = ['_static']

html_theme = "sphinx_rtd_theme"

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     '%s.tex' % project,
     u'%s Documentation' % project,
     u'OpenStack Foundation', 'manual'),
]

# Example configuration for intersphinx: refer to the Python standard library.
#intersphinx_mapping = {'http://docs.python.org/': None}

# -- Options for oslo_config.sphinxconfiggen ---------------------------------

_config_generator_config_files = [
    'ml2_nexus.ini',
    'ml2_nexus_vxlan_type_driver.ini',
    'ml2_ucsm.ini'
]

def _get_config_generator_config_definition(config_file):
    config_file_path = '../../etc/oslo-config-generator/%s' % config_file
    # oslo_config.sphinxconfiggen appends '.conf.sample' to the filename,
    # strip file extentension (.conf or .ini).
    output_file_path = '_static/config-samples/%s' % config_file.rsplit('.', 1)[0]
    return (config_file_path, output_file_path)


config_generator_config_file = [
    _get_config_generator_config_definition(conf)
    for conf in _config_generator_config_files
]



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# A tiny and simple parser for the configs fetched from a cisco IOS device.
#

import re


class LineItem(object):

    def __init__(self, line):
        self.line = line
        self.text = line.strip()
        self.parent = self
        self.children = None

    def add_children(self, child):
        if self.children:
            self.children.append(child)
        else:
            self.children = [child]

    def str_list(self):
        if self.children:
            return [self.line] + [i.str_list() for i in self.children]
        else:
            return self.line

    def re_search_children(self, linespec):
        return HTParser(self.str_list()).find_objects(linespec)

    def re_match(self, regex, group=1, default=""):
        match = re.match(regex, self.line)
        return match.group(group) if match else default

    def __repr__(self):
        return "<{} '{}'>".format(self.__class__.__name__, self.text)

    def __eq__(self, other):
        return self.str_list() == other.str_list()


class HTParser(object):
    """
    A simple hierarchical text parser.

    Indents in the text are used to derive parent child
    hierarchy.
    """

    def __init__(self, cfg):
        self._indent_list = []  # Stores items as (<indent_level>,<item>)
        if isinstance(cfg, list):
            self.cfg = cfg
        elif isinstance(cfg, str):
            self.cfg = [x for x in cfg.splitlines() if x]
        else:
            raise TypeError

    def _build_indent_based_list(self):
        self._indent_list = []
        for line in self.cfg:
            match = re.match(r'([ ]*)[^! ]', line)
            if match:
                item = (len(match.group(1)), line)
                self._indent_list.append(item)

    def _find_starts(self, linespec):
        """
        Finds the start points.

        Start points matching the linespec regex are returned as list in the
        following format:
        [(item, index), (item, index).....
         """
        linespec += ".*"
        start_points = []
        for item in self._indent_list:
            match = re.search(linespec, item[1])
            if match:
                entry = (item, self._indent_list.index(item))
                start_points.append(entry)
        return start_points

    def _find_next_indent_level(self, index):
        current_indent_level = self._indent_list[index][0]
        for item in self._indent_list[(index + 1):]:
            if item[0] > current_indent_level:
                return item[0]
            else:
                return None

    def find_lines(self, linespec):
        """Find lines that match the linespec regex."""
        res = []
        linespec += ".*"
        for line in self.cfg:
            match = re.search(linespec, line)
            if match:
                res.append(match.group(0))
        return res

    def find_objects(self, linespec):
        """Find lines that match the linespec regex.

        :param linespec: regular expression of line to match
        :return: list of LineItem objects
        """
        # Note(asr1kteam): In this code we are only adding children one-level
        # deep to a given parent (linespec), as that satisfies the IOS conf
        # parsing.
        # Note(asr1kteam): Not tested with tabs in the config. Currently used
        # with IOS config where we haven't seen tabs, but may be needed for a
        # more general case.
        res = []
        self._build_indent_based_list()
        for item, index in self._find_starts(linespec):
            parent = LineItem(item[1])
            next_ident_level = self._find_next_indent_level(index)
            if next_ident_level:
                # We start iterating from the next element
                for item in self._indent_list[(index + 1):]:
                    if item[0] == next_ident_level:
                        parent.add_children(LineItem(item[1]))
                    elif item[0] > next_ident_level:  # We skip higher indent
                        continue
                    else:  # Indent level is same or lesser than item
                        break
            res.append(parent)
        return res

    def find_children(self, linespec):
        """Find lines and immediate children that match the linespec regex.

        :param linespec: regular expression of line to match
        :returns: list of lines. These correspond to the lines that were
            matched and their immediate children
        """
        res = []
        for parent in self.find_objects(linespec):
            res.append(parent.line)
            res.extend([child.line for child in parent.children])
        return res



# Copyright (c) 2013-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Cisco Nexus-OS XML-based configuration snippets.
"""


# The following are standard strings, messages used to communicate with Nexus.
EXEC_CONF_SNIPPET = """
      <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <configure>
          <__XML__MODE__exec_configure>%s
          </__XML__MODE__exec_configure>
        </configure>
      </config>
"""

EXEC_GET_INTF_SNIPPET = """
    <cmd>show running-config interface %s %s</cmd>
"""

EXEC_GET_INVENTORY_SNIPPET = """
    <cmd>show inventory</cmd>
"""

EXEC_SAVE_CONF_SNIPPET = """
            <cmd>copy running-config startup-config</cmd>
"""

# 'show run vlan' was selected over 'show vlan' because the latter does not
# show vn-segment information
EXEC_GET_VLAN_SNIPPET = """
    <cmd>show run vlan</cmd>
"""

CMD_VLAN_CONF_VNSEGMENT_SNIPPET = """
            <vlan>
              <vlan-id-create-delete>
                <__XML__PARAM_value>%s</__XML__PARAM_value>
                <__XML__MODE_vlan>
                  <vn-segment>
                    <vlan-vnsegment>%s</vlan-vnsegment>
                  </vn-segment>
                </__XML__MODE_vlan>
              </vlan-id-create-delete>
            </vlan>
"""

CMD_VLAN_CREATE_SNIPPET = """
            <vlan>
              <vlan-id-create-delete>
                <__XML__PARAM_value>%s</__XML__PARAM_value>
              </vlan-id-create-delete>
            </vlan>
"""

CMD_VLAN_ACTIVE_SNIPPET = """
            <vlan>
              <vlan-id-create-delete>
                <__XML__PARAM_value>%s</__XML__PARAM_value>
                <__XML__MODE_vlan>
                  <state>
                    <vstate>active</vstate>
                  </state>
                </__XML__MODE_vlan>
              </vlan-id-create-delete>
            </vlan>
"""

CMD_VLAN_NO_SHUTDOWN_SNIPPET = """
            <vlan>
              <vlan-id-create-delete>
                <__XML__PARAM_value>%s</__XML__PARAM_value>
                <__XML__MODE_vlan>
                  <no>
                    <shutdown/>
                  </no>
                </__XML__MODE_vlan>
              </vlan-id-create-delete>
            </vlan>
"""

CMD_NO_VLAN_CONF_SNIPPET = """
          <no>
          <vlan>
            <vlan-id-create-delete>
              <__XML__PARAM_value>%s</__XML__PARAM_value>
            </vlan-id-create-delete>
          </vlan>
          </no>
"""

CMD_INT_VLAN_NATIVE_HEADER = """
                    <native>
                      <vlan>"""

CMD_INT_VLAN_NATIVE_TRAILER = """
                      </vlan>
                    </native>
"""

CMD_INT_VLAN_ALLOWED_HEADER = """
                    <allowed>
                      <vlan>"""

CMD_INT_VLAN_ALLOWED_TRAILER = """
                      </vlan>
                    </allowed>
"""

CMD_INT_VLAN_HEADER = """
          <interface>
            <%s>
              <interface>%s</interface>
              <__XML__MODE_if-ethernet-switch>
                <switchport>
                  <trunk>"""

CMD_VLAN_ID = """
                          <vlan_id>%s</vlan_id>"""

CMD_VLAN_ADD_ID = """
                        <add>%s
                        </add>""" % CMD_VLAN_ID

CMD_INT_VLAN_TRAILER = """
                  </trunk>
                </switchport>
              </__XML__MODE_if-ethernet-switch>
            </%s>
          </interface>
"""

CMD_INT_VLAN_SNIPPET = (CMD_INT_VLAN_HEADER +
                        CMD_INT_VLAN_ALLOWED_HEADER +
                        CMD_VLAN_ID +
                        CMD_INT_VLAN_ALLOWED_TRAILER +
                        CMD_INT_VLAN_TRAILER)

CMD_INT_VLAN_ADD_SNIPPET = (CMD_INT_VLAN_HEADER +
                            CMD_INT_VLAN_ALLOWED_HEADER +
                            CMD_VLAN_ADD_ID +
                            CMD_INT_VLAN_ALLOWED_TRAILER +
                            CMD_INT_VLAN_TRAILER)

CMD_INT_VLAN_NATIVE_SNIPPET = (CMD_INT_VLAN_HEADER +
                               CMD_INT_VLAN_NATIVE_HEADER +
                               CMD_VLAN_ID +
                               CMD_INT_VLAN_NATIVE_TRAILER +
                               CMD_INT_VLAN_TRAILER)


CMD_PORT_TRUNK = """
          <interface>
            <%s>
              <interface>%s</interface>
              <__XML__MODE_if-ethernet-switch>
                <switchport></switchport>
                <switchport>
                  <mode>
                    <trunk>
                    </trunk>
                  </mode>
                </switchport>
              </__XML__MODE_if-ethernet-switch>
            </%s>
          </interface>
"""

CMD_NO_SWITCHPORT = """
          <interface>
            <%s>
              <interface>%s</interface>
              <__XML__MODE_if-ethernet-switch>
                <no>
                  <switchport>
                  </switchport>
                </no>
              </__XML__MODE_if-ethernet-switch>
            </%s>
          </interface>
"""

CMD_NO_VLAN_INT_SNIPPET = """
          <interface>
            <%s>
              <interface>%s</interface>
              <__XML__MODE_if-ethernet-switch>
                <switchport></switchport>
                <switchport>
                  <trunk>
                    <allowed>
                      <vlan>
                        <remove>
                          <vlan>%s</vlan>
                        </remove>
                      </vlan>
                    </allowed>
                  </trunk>
                </switchport>
              </__XML__MODE_if-ethernet-switch>
            </%s>
          </interface>
"""

CMD_NO_VLAN_INT_NATIVE_SNIPPET = """
          <interface>
            <%s>
              <interface>%s</interface>
              <__XML__MODE_if-ethernet-switch>
                <switchport></switchport>
                <no>
                  <switchport>
                    <trunk>
                      <native>
                        <vlan>
                        </vlan>
                      </native>
                    </trunk>
                  </switchport>
                </no>
              </__XML__MODE_if-ethernet-switch>
            </%s>
          </interface>
"""
CMD_VLAN_SVI_SNIPPET = """
<interface>
    <vlan>
        <vlan>%s</vlan>
        <__XML__MODE_vlan>
            <no>
              <shutdown/>
            </no>
            <ip>
                <address>
                    <address>%s</address>
                </address>
            </ip>
        </__XML__MODE_vlan>
    </vlan>
</interface>
"""

CMD_NO_VLAN_SVI_SNIPPET = """
<no>
    <interface>
        <vlan>
            <vlan>%s</vlan>
        </vlan>
    </interface>
</no>
"""

CMD_INT_NVE_SNIPPET = """
<interface>
    <nve>nve%s</nve>
    <__XML__MODE_if-nve>
        <noshut>no shutdown</noshut>
        <srcint>source-interface loopback %s</srcint>
    </__XML__MODE_if-nve>
</interface>
"""

CMD_NO_INT_NVE_SNIPPET = """
<nonve>no interface nve %s</nonve>
"""

CMD_INT_NVE_MEMBER_SNIPPET = """
<interface>
    <nve>nve%s</nve>
    <__XML__MODE_if-nve>
        <member>member vni %s mcast-group %s</member>
    </__XML__MODE_if-nve>
</interface>
"""

CMD_INT_NVE_NO_MEMBER_SNIPPET = """
<interface>
    <nve>nve%s</nve>
    <__XML__MODE_if-nve>
        <member>no member vni %s</member>
    </__XML__MODE_if-nve>
</interface>
"""

CMD_FEATURE_VXLAN_SNIPPET = """
<feature>feature nv overlay</feature>
<feature>feature vn-segment-vlan-based</feature>
"""

# Removing the "feature nv overlay" configuration also removes the
# "interface nve" configuration.
CMD_NO_FEATURE_VXLAN_SNIPPET = """
<feature>no feature nv overlay</feature>
<feature>no feature vn-segment-vlan-based</feature>
"""

# REGEX SNIPPETS For extracting data from get calls

RE_GET_VLAN_ID = "vlanid-utf\>(\d+)\<"
RE_GET_VLAN_NAME = "vlanname\>([\x21-\x7e]+)\<"
RE_GET_VLAN_STATE = "vlanstate\>(\w+)\<"
RE_GET_VLAN_SHUT_STATE = "shutstate\>([a-z]+)\<"
